import re
import typing

# from structlog.stdlib import BoundLogger
import numpy as np
from ase.dft.kpoints import get_monkhorst_pack_size_and_offset, monkhorst_pack
from nomad.datamodel.context import Context
from nomad.datamodel.data import ArchiveSection
from nomad.metainfo import (
    URL,
    MEnum,
    Quantity,
    SubSection,
)
from nomad.units import ureg
from scipy.interpolate import UnivariateSpline

from nomad_simulations.schema_packages.atoms_state import ParticleState
from nomad_simulations.schema_packages.data_types import positive_float
from nomad_simulations.schema_packages.model_method import BaseModelMethod, ModelMethod
from nomad_simulations.schema_packages.numerical_settings import NumericalSettings

MOL = 6.022140857e23
FF_TOL = 1e-2


class ParameterEntry(ArchiveSection):
    """
    Generic section defining a parameter name and value
    """

    name = Quantity(
        type=str,
        shape=[],
        description="""
        Name of the parameter.
        """,
    )

    value = Quantity(
        type=str,
        shape=[],
        description="""
        Value of the parameter as a string.
        """,
    )

    unit = Quantity(
        type=str,
        shape=[],
        description="""
        Unit of the parameter as a string.
        """,
    )


#     # TODO add description quantity


class ForceCalculations(NumericalSettings):
    """
    Section containing the parameters describing how a ForceField model
    is evaluated during a simulation.
    """

    vdw_cutoff = Quantity(
        type=np.float64,
        shape=[],
        unit='m',
        description="""
        Cutoff for calculating VDW forces.
        """,
    )

    coulomb_type = Quantity(
        type=MEnum(
            'cutoff',
            'ewald',
            'multilevel_summation',
            'particle_mesh_ewald',
            'particle_particle_particle_mesh',
            'reaction_field',
        ),
        shape=[],
        description="""
        Method used for calculating long-ranged Coulomb forces.

        Allowed values are:

        | Method Name          | Description                               |

        | ---------------------- | ----------------------------------------- |

        | `""`                   | No thermostat               |

        | `"Cutoff"`          | Simple cutoff scheme. |

        | `"Ewald"` | Standard Ewald summation as described in any solid-state physics text. |

        | `"Multi-Level Summation"` |  D. Hardy, J.E. Stone, and K. Schulten,
        [Parallel. Comput. **35**, 164](https://doi.org/10.1016/j.parco.2008.12.005)|

        | `"Particle-Mesh-Ewald"`        | T. Darden, D. York, and L. Pedersen,
        [J. Chem. Phys. **98**, 10089 (1993)](https://doi.org/10.1063/1.464397) |

        | `"Particle-Particle Particle-Mesh"` | See e.g. Hockney and Eastwood, Computer Simulation Using Particles,
        Adam Hilger, NY (1989). |

        | `"Reaction-Field"` | J.A. Barker and R.O. Watts,
        [Mol. Phys. **26**, 789 (1973)](https://doi.org/10.1080/00268977300102101)|
        """,
    )

    coulomb_cutoff = Quantity(
        type=np.float64,
        shape=[],
        unit='m',
        description="""
        Cutoff for calculating short-ranged Coulomb forces.
        """,
    )

    neighbor_update_frequency = Quantity(
        type=int,
        shape=[],
        description="""
        Number of timesteps between updating the neighbor list.
        """,
    )

    neighbor_update_cutoff = Quantity(
        type=np.float64,
        shape=[],
        unit='m',
        description="""
        The distance cutoff for determining the neighbor list.
        """,
    )


class Potential(BaseModelMethod):
    """
    Section containing information about an interaction potential.

        name: str - potential name, can be as specific as needed
        type: str - potential type, e.g., 'bond', 'angle', 'dihedral', 'improper dihedral', 'nonbonded'
        functional_form: str - functional form of the potential, e.g., 'harmonic', 'Morse', 'Lennard-Jones'
        external_reference: URL
    """

    parameters = SubSection(
        sub_section=ParameterEntry.m_def,
        repeats=True,
        description="""
        List of parameters for custom potentials.
        """,
    )

    type = Quantity(
        type=MEnum(
            'bond',
            'angle',
            'bond-angle',
            'dihedral',
            'angle-dihedral',
            'improper dihedral',
            'nonbonded',
        ),
        shape=[],
        description="""
        Denotes the classification of the interaction.
        """,
    )

    # ? Use Enum here as well?
    functional_form = Quantity(
        type=str,
        shape=[],
        description="""
        Specifies the functional form of the interaction potential, e.g., harmonic, Morse, Lennard-Jones, etc.
        """,
    )

    n_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description="""
        Total number of interactions in the system for this potential.
        """,
    )

    n_particles = Quantity(
        type=np.int32,
        shape=[],
        description="""
        Number of particles interacting via (each instance of) this potential.
        """,
    )

    particle_labels = Quantity(
        type=np.dtype(str),
        shape=['n_interactions', 'n_particles'],
        description="""
        Labels of the particles for each instance of this potential, stored as a list of tuples.
        """,
    )

    particle_indices = Quantity(
        type=np.int32,
        shape=['n_interactions', 'n_particles'],
        description="""
        Indices of the particles for each instance of this potential, stored as a list of tuples.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        # set the dimensions based on the particle indices, if stored
        if not self.n_interactions:
            self.n_interactions = (
                len(self.particle_indices)
                if self.particle_indices is not None
                else None
            )
        if not self.n_particles:
            self.n_particles = (
                len(self.particle_indices[0])
                if self.particle_indices is not None
                else None
            )

        # check the consistency of the dimensions of the particle indices and labels
        if self.n_interactions and self.n_particles:
            if self.particle_indices is not None:
                assert len(self.particle_indices) == self.n_interactions
                assert len(self.particle_indices[0]) == self.n_particles
            if self.particle_labels is not None:
                assert len(self.particle_labels) == self.n_interactions
                assert len(self.particle_labels[0]) == self.n_particles


class TabulatedPotential(Potential):
    """
    Abstract class for tabulated potentials. The value of the potential and/or force
    is stored for a set of corresponding bin distances. The units for bins and forces
    should be set in the individual subclasses.
    """

    bins = Quantity(
        type=np.float64,
        shape=['*'],
        description="""
        List of bin angles.
        """,
    )

    energies = Quantity(
        type=np.float64,
        unit='J',
        shape=['*'],
        description="""
        List of energy values associated with each bin.
        """,
    )

    forces = Quantity(
        type=np.float64,
        shape=['*'],
        description="""
        List of force values associated with each bin.
        """,
    )

    def compute_forces(self, bins, energies, smoothing_factor=None):
        if smoothing_factor is None:
            smoothing_factor = len(bins) - np.sqrt(2 * len(bins))

        spline = UnivariateSpline(bins, energies, s=smoothing_factor)
        forces = -1.0 * spline.derivative()(bins)

        return forces

    def compute_energies(self, bins, forces, smoothing_factor=None):
        if smoothing_factor is None:
            smoothing_factor = len(bins) - np.sqrt(2 * len(bins))

        spline = UnivariateSpline(bins, forces, s=smoothing_factor)
        energies = -1.0 * np.array([spline.integral(bins[0], x) for x in bins])
        energies -= np.min(energies)

        return energies

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'TabulatedPotential'
        if not self.functional_form:
            self.functional_form = 'tabulated'
        elif self.functional_form != 'tabulated':
            logger.warning('Incorrect functional form set for %s.', self.name)

        if self.bins is not None and self.energies is not None:
            if len(self.bins) != len(self.energies):
                logger.error(
                    'bins and energies values have different length in %s', self.name
                )
        if self.bins is not None and self.forces is not None:
            if len(self.bins) != len(self.forces):
                logger.error('bins and forces have different length in %s', self.name)

        if self.bins is not None:
            smoothing_factor = len(self.bins) - np.sqrt(2 * len(self.bins))
            if self.forces is None and self.energies is not None:
                if not isinstance(self.bins, ureg.Quantity) or not isinstance(
                    self.energies, ureg.Quantity
                ):
                    logger.warning(
                        'Unable to derive tabulated forces from energies, '
                        'bins or energies do not have units.'
                    )
                    return

                try:
                    # generate forces from energies numerically using spline
                    self.forces = (
                        self.compute_forces(
                            self.bins.magnitude,
                            self.energies.magnitude,
                            smoothing_factor=smoothing_factor,
                        )
                        * self.energies.units
                        / self.bins.units
                    )
                    # re-derive energies to check consistency of the forces
                    energies = (
                        self.compute_energies(
                            self.bins.magnitude,
                            self.forces.magnitude,
                            smoothing_factor=smoothing_factor,
                        )
                        * self.energies.units
                    )

                    energies_diff = energies.to('kJ').magnitude * MOL - (
                        self.energies.to('kJ').magnitude * MOL
                        - np.min(self.energies.to('kJ').magnitude * MOL)
                    )
                    if np.all([np.abs(x) < FF_TOL for x in energies_diff]):
                        logger.warning(
                            'Tabulated forces were generated from energies in %s,'
                            'with consistency errors less than tol=%s. ',
                            self.name,
                            FF_TOL,
                        )
                    else:
                        logger.warning(
                            'Unable to derive tabulated forces from energies in %s,'
                            'consistency errors were greater than tol=%s.',
                            self.name,
                            FF_TOL,
                        )
                        self.forces = None
                except ValueError as e:
                    logger.warning(
                        'Unable to derive tabulated forces from energies in %s,'
                        'Unknown error occurred in derivation: %s',
                        self.name,
                        e,
                    )

            if self.forces is not None and self.energies is None:
                if not isinstance(self.bins, ureg.Quantity) or not isinstance(
                    self.forces, ureg.Quantity
                ):
                    logger.warning(
                        'Unable to derive tabulated energies from forces, '
                        'bins or forces do not have units.'
                    )
                    return
                try:
                    # generated energies from forces numerically using spline
                    self.energies = self.compute_energies(
                        self.bins.magnitude,
                        self.forces.magnitude,
                        smoothing_factor=smoothing_factor,
                    )
                    # re-derive forces to check consistency of the energies
                    forces = (
                        self.compute_forces(
                            self.bins.magnitude,
                            self.energies.magnitude,
                            smoothing_factor=smoothing_factor,
                        )
                        * self.forces.units
                    )

                    forces_diff = forces.to(f'kJ/{self.bins.units}').magnitude * MOL - (
                        self.forces.to(f'kJ/{self.bins.units}').magnitude * MOL
                    )
                    if np.all([np.abs(x) < FF_TOL for x in forces_diff]):
                        logger.warning(
                            'Tabulated energies were generated from forces in %s,'
                            'with consistency errors less than tol=%s. ',
                            self.name,
                            FF_TOL,
                        )
                    else:
                        logger.warning(
                            'Unable to derive tabulated energies from forces in %s,'
                            'consistency errors were greater than tol=%s.',
                            self.name,
                            FF_TOL,
                        )
                        self.energies = None
                except ValueError as e:
                    logger.warning(
                        'Unable to derive tabulated energies from forces in %s,'
                        'Unknown error occurred in derivation: %s',
                        self.name,
                        e,
                    )


class PolynomialForceConstant(ParameterEntry):
    """
    Section defining a force constant for a potential of polynomial form.
    """

    name = Quantity(
        type=str,
        shape=[],
        description="""
        Name of the force constant.
        """,
    )

    exponent = Quantity(
        type=np.int32,
        shape=[],
        description="""
        Exponent for this term in the polynomial.
        """,
    )

    value = Quantity(
        type=np.float64,
        shape=[],
        description="""
        Value of the force constant.
        """,
    )


class PolynomialPotential(Potential):
    r"""
    Abstract class for potentials with polynomial form:
    $V(x) = [\left k_1 (x - x_0) + k_2 (x - x_0)^2 + x_3 (x - x_0)^3 + \dots + C$,
    where $\{x_1, x_2, x_3 \dots}$ are the `force_constants` for each term in the polynomial
    expression and $x_0$ is the `equilibrium_value` of $x$. $C$ is an arbitrary constant (not stored).
    This class is intended to be used with another class specifying the potential type, e.g., BondPotential, AnglePotential, etc.
    """

    force_constants = SubSection(
        sub_section=PolynomialForceConstant.m_def,
        repeats=True,
        description="""
        List of force constants value and corresponding unit for polynomial potentials.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'PolynomialPotential'
        if not self.functional_form:
            self.functional_form = 'polynomial'
        elif self.functional_form != 'polynomial':
            logger.warning('Incorrect functional form set for PolynomialPotential.')


class BondPotential(Potential):
    """
    Section containing information about bond potentials.

    Suggested types are: harmonic, cubic, polynomial, Morse, fene, tabulated
    """

    equilibrium_value = Quantity(
        type=np.float64,
        unit='m',
        shape=[],
        description="""
        Specifies the equilibrium bond distance.
        """,
    )

    force_constant = Quantity(
        type=np.float64,
        shape=[],
        unit='J / m **2',
        description="""
        Specifies the force constant of the bond potential.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        if not self.name:
            self.name = 'BondPotential'
        if not self.type:
            self.type = 'bond'
        elif self.type != 'bond':
            logger.warning('Incorrect type set for BondPotential.')

        if self.n_particles:
            if self.n_particles != 2:
                logger.warning('Incorrect number of particles set for BondPotential.')
            else:
                self.n_particles = 2


class HarmonicBond(BondPotential):
    r"""
    Section containing information about a Harmonic bond potential:
    $V(r) = \frac{1}{2} k_r (r - r_0)^2 + C$,
    where $k_r$ is the `force_constant` and $r_0$ is the `equilibrium_value` of $r$.
    $C$ is an arbitrary constant (not stored).
    """

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'HarmonicBond'
        if not self.functional_form:
            self.functional_form = 'harmonic'
        elif self.functional_form != 'harmonic':
            logger.warning('Incorrect functional form set for HarmonicBond.')


class CubicBond(BondPotential):
    r"""
    Section containing information about a Cubic bond potential:
    $V(r) = \frac{1}{2} k_r (r - r_0)^2 + \frac{1}{3} k_c (r - r_0)^3 + C$,
    where $k_r$ is the (harmonic) `force_constant`, $k_c$ is the `force_constant_cubic`,
    and $r_0$ is the `equilibrium_value` of $r$.
    C is an arbitrary constant (not stored).
    """

    force_constant_cubic = Quantity(
        type=np.float64,
        shape=[],
        unit='J / m**3',
        description="""
        Specifies the cubic force constant of the bond potential.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'CubicBond'
        if not self.functional_form:
            self.functional_form = 'cubic'
        elif self.functional_form != 'cubic':
            logger.warning('Incorrect functional form set for CubicBond.')


class PolynomialBond(PolynomialPotential, BondPotential):
    """
    Section containing information about a polynomial bond potential:
    """

    def __init__(self):
        super().__init__()
        docstring = PolynomialPotential.__doc__
        pattern = r'\$V\(x\)(.*?)(\(not stored\)\.)'
        match = re.search(pattern, docstring, re.DOTALL)
        extracted_text = '<functional form missing>'
        if match:
            extracted_text = match.group().strip()  # .group(1).strip()
        self.__doc__ = rf"""{self.__doc__} {extracted_text}.
            Here the dependent variable of the potential, $x$, corresponds to the bond distance."""

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'PolynomialBond'


class MorseBond(BondPotential):
    r"""
    Section containing information about a Morse potential:
    $V(r) = D \left[ 1 - e^{-a(r - r_0)} \right]^2 + C$,
    where $a = sqrt(k/2D)$ is the `well_steepness`, with `force constant` k.
    D is the `well_depth`, and r_0 is the `equilibrium_value` of a.
    C is an arbitrary constant (not stored).
    """

    well_depth = Quantity(
        type=np.float64,
        unit='J',
        shape=[],
        description="""
        Specifies the depth of the potential well.
        """,
    )

    well_steepness = Quantity(
        type=np.float64,
        unit='1/m',
        shape=[],
        description="""
        Specifies the steepness of the potential well.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'MorseBond'
        if not self.functional_form:
            self.functional_form = 'morse'
        elif self.functional_form != 'morse':
            logger.warning('Incorrect functional form set for MorseBond.')

        if self.well_depth is not None and self.well_steepness is not None:
            self.force_constant = 2.0 * self.well_depth * self.well_steepness**2
        elif self.well_depth is not None and self.force_constant is not None:
            self.well_steepness = np.sqrt(self.force_constant / (2.0 * self.well_depth))


class FeneBond(BondPotential):
    r"""
    Section containing information about a FENE potential:
    $V(r) = -\frac{1}{2} k R_0^2 \ln \left[ 1 - \left( \frac{r - r_0}{R_0} \right)^2 \right] + C$,
    $k$ is the `force_constant`, $r_0$ is the `equilibrium_value` of $r$, and $R_0$ is the
    maximum allowable bond extension beyond $r_0$. $C$ is an arbitrary constant (not stored).
    """

    maximum_extension = Quantity(
        type=np.float64,
        unit='m',
        shape=[],
        description="""
        Specifies the maximum extension beyond the equilibrium bond distance.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'FeneBond'
        if not self.functional_form:
            self.functional_form = 'fene'
        elif self.functional_form != 'fene':
            logger.warning('Incorrect functional form set for FeneBond.')


class TabulatedBond(TabulatedPotential, BondPotential):
    """
    Section containing information about a tabulated bond potential.
    The value of the potential and/or force is stored for a set of corresponding bin distances.
    """

    bins = Quantity(
        type=np.float64,
        unit='m',
        shape=['*'],
        description="""
        List of bin distances.
        """,
    )

    forces = Quantity(
        type=np.float64,
        unit='J/m',
        shape=['*'],
        description="""
        List of force values associated with each bin.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'TabulatedBond'


class AnglePotential(Potential):
    """
    Section containing information about angle potentials.

    Suggested types are: harmonic, cosine, restricted_cosine, fourier_series, urey_bradley, polynomial, tabulated
    """

    equilibrium_value = Quantity(
        type=np.float64,
        unit='radian',
        shape=[],
        description="""
        Specifies the equilibrium angle.
        """,
    )

    force_constant = Quantity(
        type=np.float64,
        shape=[],
        unit='J / radian**2',
        description="""
        Specifies the force constant of the angle potential.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'AnglePotential'
        if not self.type:
            self.type = 'angle'
        elif self.type != 'angle':
            logger.warning('Incorrect type set for AnglePotential.')

        if self.n_particles:
            if self.n_particles != 3:
                logger.warning('Incorrect number of particles set for AnglePotential.')
            else:
                self.n_particles = 3


class LinearAngle(AnglePotential):
    r"""
    Section containing information about a linear angle potential:
    $V(\theta) = k_\theta (\theta - \theta_0) + C$,
    where $k_\theta$ is the `force_constant` and $\theta_0$ is the `equilibrium_value`.
    $C$ is an arbitrary constant (not stored).
    """

    force_constant = Quantity(
        type=np.float64,
        shape=[],
        unit='J / radian',
        description="""
        Specifies the force constant of the angle potential.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'LinearAngle'
        if not self.functional_form:
            self.functional_form = 'linear'
        elif self.functional_form != 'linear':
            logger.warning('Incorrect functional form set for LinearAngle.')


class HarmonicAngle(AnglePotential):
    r"""
    Section containing information about a Harmonic angle potential:
    $V(\theta) = \frac{1}{2} k_\theta (\theta - \theta_0)^2 + C$,
    where $k_\theta$ is the `force_constant` and $\theta_0$ is the `equilibrium_value` of
    the angle $\theta$. $C$ is an arbitrary constant (not stored).
    """

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'HarmonicAngle'
        if not self.functional_form:
            self.functional_form = 'harmonic'
        elif self.functional_form != 'harmonic':
            logger.warning('Incorrect functional form set for HarmonicAngle.')


class CosineAngle(AnglePotential):
    r"""
    Section containing information about a Cosine angle potential:
    $V(\theta) = \frac{1}{2} k_\theta \left[ \cos(\theta) - \cos(\theta_0) \right]^2 + C$,
    where $k_\theta$ is the `force_constant` and $\theta_0$ is the `equilibrium_value` of
    the angle $\theta$. $C$ is an arbitrary constant (not stored).

    """

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'CosineAngle'
        if not self.functional_form:
            self.functional_form = 'cosine'
        elif self.functional_form != 'cosine':
            logger.warning('Incorrect functional form set for CosineAngle.')


class RestrictedCosineAngle(AnglePotential):
    r"""
    Section containing information about a Restricted Cosine angle potential:
    $V(\theta) = \frac{1}{2} k_\theta \frac{\left[ \cos(\theta) - \cos(\theta_0) \right]^2}{\sin^2(\theta)} + C$,
    where $k_\theta$ is the `force_constant` and $\theta_0$ is the `equilibrium_value` of
    the angle $\theta$. $C$ is an arbitrary constant (not stored).

    """

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'RestrictedCosineAngle'
        if not self.functional_form:
            self.functional_form = 'restricted_cosine'
        elif self.functional_form != 'restricted_cosine':
            logger.warning('Incorrect functional form set for RestrictedCosineAngle.')


# TODO I think we should name these more generally and then say that AKA
class UreyBradleyAngle(AnglePotential):
    r"""
    Section containing information about a Urey-Bradley angle potential:
    $V(\theta) = \frac{1}{2} k_\theta (\theta - \theta_0)^2 +
    \frac{1}{2} k_{13} (r_{13} - r_{13}^0)^2 + C$,
    where where $k_\theta$ is the `force_constant` and $\theta_0$ is the `equilibrium_value` of
    the angle $\theta$, as for a harmonic angle potential. $k_{13}$ is the `force_constant_UB`
    for the 1-3 term, and $r_{13}^0$ is the `equilibrium_value_UB` of the 1-3 distance (i.e.,
    the distance between the edge particles forming the 1-2-3 angle, $\theta$), $r_{13}$.
    $C$ is an arbitrary constant (not stored).
    """

    equilibrium_value_UB = Quantity(
        type=np.float64,
        unit='m',
        shape=[],
        description="""
        Specifies the equilibrium 1-3 distance.
        """,
    )

    force_constant_UB = Quantity(
        type=np.float64,
        shape=[],
        unit='J / m **2',
        description="""
        Specifies the force constant of the potential.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'UreyBradleyAngle'
        if not self.functional_form:
            self.functional_form = 'urey_bradley'
        elif self.functional_form != 'urey_bradley':
            logger.warning('Incorrect functional form set for UreyBradleyAngle.')


class PolynomialAngle(PolynomialPotential, AnglePotential):
    """
    Section containing information about a polynomial angle potential:
    """

    def __init__(self):
        super().__init__()
        docstring = PolynomialPotential.__doc__
        pattern = r'\$V\(x\)(.*?)(\(not stored\)\.)'
        match = re.search(pattern, docstring, re.DOTALL)
        extracted_text = '<functional form missing>'
        if match:
            extracted_text = match.group().strip()  # .group(1).strip()
        self.__doc__ = rf"""{self.__doc__} {extracted_text}.
            Here the dependent variable of the potential, $x$, corresponds to the angle
            between three particles."""

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'PolynomialAngle'


class TabulatedAngle(AnglePotential, TabulatedPotential):
    """
    Section containing information about a tabulated bond potential. The value of the
    potential and/or force
    is stored for a set of corresponding bin distances.
    """

    bins = Quantity(
        type=np.float64,
        unit='radian',
        shape=['*'],
        description="""
        List of bin angles.
        """,
    )

    forces = Quantity(
        type=np.float64,
        unit='J/radian',
        shape=['*'],
        description="""
        List of force values associated with each bin.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'TabulatedAngle'


class BondAngleCouplingPotential(Potential):
    """
    Section containing information about bond potentials.

    Suggested types are: linear
    """

    equilibrium_bond_length = Quantity(
        type=np.float64,
        unit='m',
        shape=[],
        description="""
        Equilibrium bond length $r_0$.
        """,
    )

    equilibrium_angle = Quantity(
        type=np.float64,
        unit='radian',
        shape=[],
        description="""
        Equilibrium angle $\theta_0$.
        """,
    )

    force_constant = Quantity(
        type=np.float64,
        unit='J / (m * radian)',
        shape=[],
        description="""
        Force constant coupling bond length and angle.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'BondAngleCouplingPotential'
        if not self.type:
            self.type = 'bond-angle'
        elif self.type != 'bond-angle':
            logger.warning('Incorrect type set for BondAngleCouplingPotential.')

        if self.n_particles:
            if self.n_particles != 3:
                logger.warning(
                    'Incorrect number of particles set for BondAngleCouplingPotential.'
                )
            else:
                self.n_particles = 3


class LinearBondAngleCoupling(BondAngleCouplingPotential):
    r"""
    Section containing a linear bond–angle coupling potential:
    $V(r, \theta) = k (r - r_0)(\theta - \theta_0) + C$,
    where $r_0$ and $\theta_0$ are the equilibrium values.
    """

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'LinearBondAngleCoupling'
        if not self.functional_form:
            self.functional_form = 'linear'
        elif self.functional_form != 'linear':
            logger.warning('Incorrect functional form set for LinearBondAngleCoupling.')


class DihedralPotential(Potential):
    """
    Section containing information about dihedral potentials.

    Suggested types are: fourier_series, tabulated

    # ? Something about angle convention?
    """

    equilibrium_value = Quantity(
        type=np.float64,
        unit='radian',
        shape=[],
        description="""
        Specifies the equilibrium dihedral angle.
        """,
    )

    force_constant = Quantity(
        type=np.float64,
        shape=[],
        unit='J / radian**2',
        description="""
        Specifies the force constant of the dihedral angle potential.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'DihedralPotential'
        if not self.type:
            self.type = 'dihedral'
        elif self.type != 'dihedral':
            logger.warning('Incorrect type set for DihedralPotential.')

        if self.n_particles:
            if self.n_particles != 4:
                logger.warning(
                    'Incorrect number of particles set for DihedralPotential.'
                )
            else:
                self.n_particles = 4


class PeriodicDihedral(DihedralPotential):
    r"""
    Section for periodic proper dihedral potential:
    $V(\phi) = \frac{1}{2} k [1 + \cos(n \phi - \delta)] + C$
    """

    multiplicity = Quantity(
        type=np.int32,
        shape=[],
        description="""
        Periodicity $n$.
        """,
    )

    phase_shift = Quantity(
        type=np.float64,
        unit='radian',
        shape=[],
        description=r"""
        Phase shift $\delta$.
        """,
    )

    force_constant = Quantity(
        type=np.float64,
        unit='J',
        shape=[],
        description="""
        Amplitude $k$.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'PeriodicDihedral'
        if not self.functional_form:
            self.functional_form = 'periodic'


class RyckaertBellemansDihedral(DihedralPotential):
    r"""
    Ryckaert-Bellemans dihedral:
    $V(\phi) = \sum_{n=0}^{5} C_n \cos^n(\phi)$
    """

    coefficients = Quantity(
        type=np.float64,
        shape=[6],
        unit='J',
        description="""
        Coefficients $C_0$ through $C_5$.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'RyckaertBellemansDihedral'
        if not self.functional_form:
            self.functional_form = 'ryckaert_bellemans'


class TabulatedDihedral(DihedralPotential, TabulatedPotential):
    """
    Section containing information about a tabulated bond potential. The value of the potential and/or force
    is stored for a set of corresponding bin distances.
    """

    bins = Quantity(
        type=np.float64,
        unit='radian',
        shape=['*'],
        description="""
        List of bin dihedral angles.
        """,
    )

    forces = Quantity(
        type=np.float64,
        unit='J/radian',
        shape=['*'],
        description="""
        List of force values associated with each bin.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'TabulatedDihedral'


class AngleDihedralCouplingPotential(Potential):
    """
    Section containing information about angle-dihedral potentials.

    Suggested types are: linear
    """

    equilibrium_angle = Quantity(type=np.float64, unit='radian', shape=[])
    equilibrium_dihedral = Quantity(type=np.float64, unit='radian', shape=[])
    force_constant_angle = Quantity(type=np.float64, unit='J/radian**2', shape=[])
    force_constant_dihedral = Quantity(type=np.float64, unit='J/radian**2', shape=[])
    coupling_constant = Quantity(type=np.float64, unit='J/radian**2', shape=[])

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'AngleDihedralCouplingPotential'
        if not self.type:
            self.type = 'angle-dihedral'
        elif self.type != 'angle-dihedral':
            logger.warning('Incorrect type set for AngleDihedralCouplingPotential.')

        if self.n_particles:
            if self.n_particles != 4:
                logger.warning(
                    'Incorrect number of particles set for AngleDihedralCouplingPotential.'
                )
            else:
                self.n_particles = 4


class HarmonicAngleDihedralCoupling(AngleDihedralCouplingPotential):
    r"""
    Harmonic form of an angle–dihedral coupling potential:
    $V = k_1(\theta - \theta_0)^2 + k_2(\phi - \phi_0)^2 + k_3(\theta - \theta_0)(\phi - \phi_0)$
    """

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'HarmonicAngleDihedralCoupling'
        if not self.functional_form:
            self.functional_form = 'angle_dihedral_coupled'
        elif self.functional_form != 'angle_dihedral_coupled':
            logger.warning(
                'Incorrect functional form set for HarmonicAngleDihedralCoupling.'
            )


class ImproperDihedralPotential(DihedralPotential):
    """
    Section containing information about improper dihedral potentials.

    Suggested types are: ...

    # ? Something about angle convention?
    """

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)
        self.name = 'ImproperDihedralPotential'
        if self.type != 'improper dihedral':
            self.type = 'improper dihedral'


class HarmonicImproper(ImproperDihedralPotential):
    r"""
    Section for harmonic improper dihedral potential:
    $V(\omega) = \frac{1}{2} k (\omega - \omega_0)^2 + C$
    """

    equilibrium_value = Quantity(
        type=np.float64,
        unit='radian',
        shape=[],
        description=r"""
        Equilibrium improper angle $\omega_0$.
        """,
    )

    force_constant = Quantity(
        type=np.float64,
        unit='J / radian**2',
        shape=[],
        description="""
        Force constant $k$.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'HarmonicImproper'
        if not self.functional_form:
            self.functional_form = 'harmonic'


class PeriodicImproper(ImproperDihedralPotential):
    r"""
    Section for periodic improper dihedral potential:
    $V(\omega) = \frac{1}{2} k [1 + \cos(n \omega - \delta)] + C$
    """

    multiplicity = Quantity(
        type=np.int32,
        shape=[],
        description="""
        Periodicity $n$.
        """,
    )

    phase_shift = Quantity(
        type=np.float64,
        unit='radian',
        shape=[],
        description=r"""
        Phase shift $\delta$.
        """,
    )

    force_constant = Quantity(
        type=np.float64,
        unit='J',
        shape=[],
        description="""
        Amplitude $k$.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'PeriodicImproper'
        if not self.functional_form:
            self.functional_form = 'periodic'


class ParticleParameters(ArchiveSection):
    """
    Force field parameters for one atom type. `species_scope` holds forward
    references from this entry to every `ParticleState` instance assigned this
    type, mirroring the pattern of `BasisSetComponent.species_scope`.
    `species_scope` is resolved automatically by `ParticleParametersContainer.normalize()`
    by matching `particle_type` against `ParticleState.label`.
    """

    species_scope = Quantity(
        type=ParticleState,
        shape=['*'],
        description="""
        References to the `ParticleState` instances assigned this particle type.
        Resolved by `ParticleParametersContainer.normalize()` from `particle_type`
        matching against `ParticleState.label` in the representative `ModelSystem`.
        """,
    )

    particle_type = Quantity(
        type=str,
        description="""
        Force field particle type label as defined by the force field (e.g. `'CT'`,
        `'OW'`, `'HW'` in AMBER/CHARMM). Multiple particles in `species_scope` may
        share this label.
        This identifier is part of the force-field parameterization, not chemical
        element/species identity.
        """,
    )

    partial_charge = Quantity(
        type=np.float64,
        unit='elementary_charge',
        description="""
        Partial charge assigned in this `AtomParameters` entry for the referenced
        `ParticleState` instance, as a force field parameter. Adjusted from formal/
        oxidation charges to model electrostatic interactions within the context
        of the force field. Distinct from formal or oxidation-state style charges.
        """,
    )

    effective_mass = Quantity(
        type=positive_float(),
        unit='kg',
        description="""
        Effective mass assigned in this `AtomParameters` entry for the referenced
        `ParticleState` instance, as a force field parameter. May differ from the
        standard atomic mass when force-field parameterization adjusts masses for
        numerical stability or coarse-grained representations.
        """,
    )


class ParticleParametersContainer(NumericalSettings):
    """
    Numerical-settings container for per-particle force field parameters grouped
    by particle type. Each `ParticleParameters` subsection covers all particles of one
    FF particle type via `species_scope` (references to `ParticleState` instances).
    """

    particle_parameters = SubSection(
        sub_section=ParticleParameters.m_def,
        repeats=True,
        description="""
        Per-particle-type force field parameters (partial charge, effective mass,
        particle type label). Each entry references its particles via `species_scope`.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)
        try:
            particle_states = archive.data.model_system[-1].particle_states
        except (AttributeError, IndexError, TypeError):
            logger.error(
                'Unable to resolve ParticleParameters.species_scope references, '
                'model_system or particle_states not accessible.'
            )
            particle_states = None

        if particle_states is not None:
            label_index: dict[str, list] = {}
            for ps in particle_states:
                lbl = getattr(ps, 'label', None)
                if lbl is not None:
                    label_index.setdefault(lbl, []).append(ps)
            for ap in self.particle_parameters or []:
                if ap.species_scope is not None and len(ap.species_scope) > 0:
                    continue
                if ap.particle_type is None:
                    continue
                ap.species_scope = label_index.get(ap.particle_type, [])

        seen: set[int] = set()
        duplicates: set[str] = set()
        for ap in self.particle_parameters or []:
            for ps in ap.species_scope or []:
                ps_id = id(ps)
                if ps_id in seen:
                    duplicates.add(str(getattr(ap, 'particle_type', None)))
                seen.add(ps_id)
        if duplicates:
            logger.error(
                'Overlapping species_scope references across ParticleParameters entries.'
                '(conflicting particle_type(s): %s)',
                ', '.join(sorted(duplicates)),
            )


class ForceField(ModelMethod):
    """
    Section containing the parameters of a (classical, particle-based) force field model.
    Typical `numerical_settings` are ForceCalculations and ParticleParametersContainer.
    Lists of interactions by type and, if available, corresponding parameters can be given
    within `interactions`.
    Additionally, a published model can be referenced with `reference`.
    """

    # name and external reference already defined in BaseModelMethod
    # name = Quantity(
    #     type=str,
    #     shape=[],
    #     description="""
    #     Identifies the name of the model.
    #     """,
    # )

    # reference = Quantity(
    #     type=str,
    #     shape=['0..*'],
    #     description="""
    #     List of references to the model e.g. DOI, URL.
    #     """,
    # )

    kimid = Quantity(
        type=URL,
        description="""
        Reference to a model stored on the OpenKim database.
        """,
    )

    #     interactions = SubSection(sub_section=Interactions.m_def, repeats=True)
    contributions = SubSection(
        sub_section=Potential.m_def,
        repeats=True,
        description="""
        Contribution or sub-term of the total model Hamiltonian.
        """,
    )

    def normalize(self, archive, logger) -> None:
        super().normalize(archive, logger)

        self.name = 'ForceField'


# TODO Need to survey Lammps and maybe openmm and check for other common potential types
# TODO prevent the base classes from being used directly as they won't pass norm
# without units



from nomad_simulations.schema_packages.workflow.gw import (
    DFTGWMethod,
    DFTGWResults,
    DFTGWWorkflow,
)
from nomad_simulations.schema_packages.workflow.single_point import SinglePoint


class TestDFTGWWorkflow:
    def test_inputs_outputs(self, logger, archive, log_output):
        workflow = DFTGWWorkflow()
        workflow.normalize(archive, logger)
        assert isinstance(workflow.method, DFTGWMethod)
        assert isinstance(workflow.results, DFTGWResults)
        assert len(workflow.inputs) == 1
        assert len(workflow.outputs) == 1
        assert workflow.inputs[0].name == 'DFT+GW workflow parameters'
        assert workflow.outputs[0].name == 'DFT+GW workflow results'
        assert 'Incorrect number of tasks found.' in log_output.entries[0]['event']

    def test_tasks(self, logger, archive):
        workflow = DFTGWWorkflow(tasks=[SinglePoint(), SinglePoint()])
        workflow.normalize(archive, logger)
        assert workflow.tasks[0].name == 'DFT'
        assert workflow.tasks[1].name == 'GW'
        assert workflow.method in [inp.section for inp in workflow.tasks[0].inputs]
        assert workflow.tasks[0] in [inp.section for inp in workflow.tasks[1].inputs]



from typing import Optional, Union

import pytest
from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.properties import (
    QuasiparticleWeight,
)
from nomad_simulations.schema_packages.properties.greens_function import (
    BaseGreensFunction,
)
from nomad_simulations.schema_packages.variables import (
    Frequency,
    ImaginaryTime,
    KMesh,
    MatsubaraFrequency,
    Time,
    WignerSeitz,
)

from . import logger


def set_variables(
    gfs: BaseGreensFunction,
    variables: list[
        WignerSeitz | KMesh | Time | ImaginaryTime | Frequency | MatsubaraFrequency
    ],
) -> BaseGreensFunction:
    """
    Set the variables for the `BaseGreensFunction` class.
    """
    for variable in variables:
        if isinstance(variable, WignerSeitz):
            gfs.wigner_seitz = variable
        elif isinstance(variable, KMesh):
            gfs.k_mesh = variable
        elif isinstance(variable, Time):
            gfs.time = variable
        elif isinstance(variable, ImaginaryTime):
            gfs.imaginary_time = variable
        elif isinstance(variable, Frequency):
            gfs.real_frequency = variable
        elif isinstance(variable, MatsubaraFrequency):
            gfs.matsubara_frequency = variable
    return gfs


class TestBaseGreensFunction:
    """
    Test the `BaseGreensFunction` class defined in `properties/greens_function.py`.
    """

    @pytest.mark.parametrize(
        'variables, result',
        [
            ([], None),
            ([WignerSeitz()], 'r'),
            ([KMesh()], 'k'),
            ([Time()], 't'),
            ([ImaginaryTime()], 'it'),
            ([Frequency()], 'w'),
            ([MatsubaraFrequency()], 'iw'),
            ([WignerSeitz(), Time()], 'rt'),
            ([WignerSeitz(), ImaginaryTime()], 'rit'),
            ([WignerSeitz(), Frequency()], 'rw'),
            ([WignerSeitz(), MatsubaraFrequency()], 'riw'),
            ([KMesh(), Time()], 'kt'),
            ([KMesh(), ImaginaryTime()], 'kit'),
            ([KMesh(), Frequency()], 'kw'),
            ([KMesh(), MatsubaraFrequency()], 'kiw'),
        ],
    )
    def test_resolve_space_id(
        self,
        variables: list[
            WignerSeitz | KMesh | Time | ImaginaryTime | Frequency | MatsubaraFrequency
        ],
        result: str,
    ):
        """
        Test the `resolve_space_id` method of the `BaseGreensFunction` class.
        """
        gfs = BaseGreensFunction(n_atoms=1, n_correlated_orbitals=1)
        gfs = set_variables(gfs, variables)
        assert gfs.resolve_space_id() == result

    @pytest.mark.parametrize(
        'space_id, variables, result',
        [
            ('', [], None),  # empty `space_id`
            ('rt', [], None),  # `space_id` set by parser
            ('', [WignerSeitz()], 'r'),  # resolving `space_id`
            ('rt', [WignerSeitz()], 'r'),  # normalize overwrites `space_id`
            ('', [KMesh()], 'k'),
            ('', [Time()], 't'),
            ('', [ImaginaryTime()], 'it'),
            ('', [Frequency()], 'w'),
            ('', [MatsubaraFrequency()], 'iw'),
            ('', [WignerSeitz(), Time()], 'rt'),
            ('', [WignerSeitz(), ImaginaryTime()], 'rit'),
            ('', [WignerSeitz(), Frequency()], 'rw'),
            ('', [WignerSeitz(), MatsubaraFrequency()], 'riw'),
            ('', [KMesh(), Time()], 'kt'),
            ('', [KMesh(), ImaginaryTime()], 'kit'),
            ('', [KMesh(), Frequency()], 'kw'),
            ('', [KMesh(), MatsubaraFrequency()], 'kiw'),
        ],
    )
    def test_normalize(
        self,
        space_id: str,
        variables: list[
            WignerSeitz | KMesh | Time | ImaginaryTime | Frequency | MatsubaraFrequency
        ],
        result: str | None,
    ):
        """
        Test the `normalize` method of the `BaseGreensFunction` class.
        """
        gfs = BaseGreensFunction(n_atoms=1, n_correlated_orbitals=1)
        gfs = set_variables(gfs, variables)
        gfs.space_id = space_id if space_id else None
        gfs.normalize(archive=EntryArchive(), logger=logger)
        assert gfs.space_id == result


class TestQuasiparticleWeight:
    """
    Test the `QuasiparticleWeight` class defined in `properties/greens_function.py`.
    """

    @pytest.mark.parametrize(
        'value, result',
        [
            ([[1, 0.9, 0.8]], 'non-correlated metal'),
            ([[0.2, 0.3, 0.1]], 'strongly-correlated metal'),
            ([[0, 0.3, 0.1]], 'OSMI'),
            ([[0, 0, 0]], 'Mott insulator'),
            ([[1.0, 0.8, 0.2]], None),
        ],
    )
    def test_resolve_system_correlation_strengths(
        self, value: list[float], result: str | None
    ):
        """
        Test the `resolve_system_correlation_strengths` method of the `QuasiparticleWeight` class.
        """
        quasiparticle_weight = QuasiparticleWeight(
            n_atoms=1, n_correlated_orbitals=3, value=value
        )
        assert quasiparticle_weight.resolve_system_correlation_strengths() == result

    @pytest.mark.parametrize(
        'value, reference',
        [
            ([[1, 0.5, -2]], ''),
            ([[1, 0.5, 8]], ''),
            ([[1, 0.9, 0.8]], 'non-correlated metal'),
            ([[0.2, 0.3, 0.1]], 'strongly-correlated metal'),
            ([[0, 0.3, 0.1]], 'OSMI'),
            ([[0, 0, 0]], 'Mott insulator'),
            # ? ([[1.0, 0.8, 0.2]], ''),
        ],
    )
    def test_normalize(self, value: list[float], reference: str):
        """
        Test the `normalize` method of the `QuasiparticleWeight` class.
        """
        if reference == '':
            with pytest.raises(ValueError, match=r'All values must be in \[0,1\]'):
                QuasiparticleWeight(n_atoms=1, n_correlated_orbitals=3, value=value)
        else:
            quasiparticle_weight = QuasiparticleWeight(
                n_atoms=1, n_correlated_orbitals=3, value=value
            )
            quasiparticle_weight.normalize(archive=EntryArchive(), logger=logger)
            assert quasiparticle_weight.system_correlation_strengths == reference



from typing import TYPE_CHECKING

import numpy as np
from nomad.metainfo import Quantity, Section

if TYPE_CHECKING:
    from nomad.datamodel.context import Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.physical_property import PhysicalProperty


class RadiusOfGyration(PhysicalProperty):
    """
    Radius of gyration (Rg).
    """

    value = Quantity(
        type=np.float64,
        shape=[],
        unit='m',
        description="""
        Value of Rg.
        """,
    )



from nomad_simulations.schema_packages.model_system import (
    ModelSystem,
    AlternativeRepresentation
)

# Create model system with original cell data
model_system = ModelSystem()
model_system.lattice_vectors = original_lattice_vectors
model_system.positions = original_positions

# Add primitive cell representation
primitive_rep = AlternativeRepresentation(
    name='primitive',
    crystal_cell_type='primitive',
    lattice_vectors=primitive_lattice_vectors,
    transformation_matrix=primitive_transformation
)
model_system.representations.append(primitive_rep)

# Add conventional cell representation
conventional_rep = AlternativeRepresentation(
    name='conventional',
    crystal_cell_type='conventional',
    lattice_vectors=conventional_lattice_vectors
)
model_system.representations.append(conventional_rep)



from nomad_simulations.schema_packages.general import Simulation
from nomad_simulations.schema_packages.model_method import DFT, XCFunctional

simulation = Simulation()
dft = DFT(name='DFT', type='KS', xc=XCFunctional(functional_key='PBE'))
simulation.model_method.append(dft)

assert isinstance(simulation.model_method[0], DFT)
assert simulation.model_method[0].xc.functional_key == 'PBE'



from typing import Optional, Union

import numpy as np
import pytest
from nomad.datamodel import EntryArchive
from nomad.units import ureg

from nomad_simulations.schema_packages.properties import ElectronicBandGap
from nomad_simulations.schema_packages.variables import Temperature

from . import logger


class TestElectronicBandGap:
    """
    Test the `ElectronicBandGap` class defined in `properties/band_gap.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    def test_default_quantities(self):
        """
        Test the default quantities assigned when creating an instance of the `ElectronicBandGap` class.
        """
        electronic_band_gap = ElectronicBandGap()
        assert (
            electronic_band_gap.iri
            == 'http://fairmat-nfdi.eu/taxonomy/ElectronicBandGap'
        )

    @pytest.mark.parametrize(
        'momentum_transfer, type, result',
        [
            (None, None, None),
            (None, 'direct', 'direct'),
            (None, 'indirect', 'indirect'),
            ([[0, 0, 0]], None, None),
            ([[0, 0, 0]], 'direct', None),
            ([[0, 0, 0]], 'indirect', None),
            ([[0, 0, 0], [0, 0, 0]], None, 'direct'),
            ([[0, 0, 0], [0, 0, 0]], 'direct', 'direct'),
            ([[0, 0, 0], [0, 0, 0]], 'indirect', 'direct'),
            ([[0, 0, 0], [0.5, 0.5, 0.5]], None, 'indirect'),
            ([[0, 0, 0], [0.5, 0.5, 0.5]], 'direct', 'indirect'),
            ([[0, 0, 0], [0.5, 0.5, 0.5]], 'indirect', 'indirect'),
        ],
    )
    def test_resolve_type(
        self, momentum_transfer: list[float] | None, type: str, result: str | None
    ):
        """
        Test the `resolve_type` method.
        """
        electronic_band_gap = ElectronicBandGap(
            momentum_transfer=momentum_transfer,
            type=type,
        )
        assert electronic_band_gap.resolve_type(logger=logger) == result

    def test_normalize(self):
        """
        Test the `normalize` method for two different ElectronicBandGap instantiations, one with a scalar
        `value` and another with a temperature-dependent `value`
        """
        scalar_band_gap = ElectronicBandGap(type='direct')
        scalar_band_gap.value = 1.0 * ureg.joule
        scalar_band_gap.normalize(EntryArchive(), logger)
        assert scalar_band_gap.type == 'direct'
        assert np.isclose(scalar_band_gap.value.magnitude, 1.0)

        # t_dependent_band_gap = ElectronicBandGap(
        #     variables=[Temperature(points=[0, 10, 20, 30] * ureg.kelvin)],
        #     type='direct',
        # )
        # t_dependent_band_gap.value = [1.0, 2.0, 3.0, 4.0] * ureg.joule
        # t_dependent_band_gap.normalize(EntryArchive(), logger)
        # assert t_dependent_band_gap.type == 'direct'
        # assert (
        #     np.isclose(t_dependent_band_gap.value.magnitude, [1.0, 2.0, 3.0, 4.0])
        # ).all()



from __future__ import annotations


def build_relation_legend(
    *, has_inherit: bool, has_contain: bool, has_refs: bool
) -> list[str]:
    """Return HTML legend items that visually track Mermaid relationship arrows."""
    items: list[str] = []

    if has_inherit:
        items.append(
            '<div class="uml-legend__item" role="listitem">'
            '<svg class="uml-legend__swatch" viewBox="0 0 64 16" aria-hidden="true">'
            '<line class="uml-legend__line" x1="54" y1="8" x2="22" y2="8"/>'
            '<path class="uml-legend__head uml-legend__head--open" d="M10 8 L22 2 L22 14 Z"/>'
            '</svg>'
            '<span>inheritance (is-a)</span>'
            '</div>'
        )

    if has_contain:
        items.append(
            '<div class="uml-legend__item" role="listitem">'
            '<svg class="uml-legend__swatch" viewBox="0 0 64 16" aria-hidden="true">'
            '<path class="uml-legend__head uml-legend__head--filled" d="M10 8 L16 2 L22 8 L16 14 Z"/>'
            '<line class="uml-legend__line" x1="22" y1="8" x2="52" y2="8"/>'
            '</svg>'
            '<span>composition (has-a)</span>'
            '</div>'
        )

    if has_refs:
        items.append(
            '<div class="uml-legend__item" role="listitem">'
            '<svg class="uml-legend__swatch" viewBox="0 0 64 16" aria-hidden="true">'
            '<line class="uml-legend__line uml-legend__line--dashed" x1="8" y1="8" x2="40" y2="8"/>'
            '<path class="uml-legend__head uml-legend__head--open" d="M40 8 L48 4 M40 8 L48 12"/>'
            '</svg>'
            '<span><code>A ..&gt; B</code> references relationship, A depends on B</span>'
            '</div>'
        )

    return items


def wrap_diagram_card(
    diagram_lines: list[str], legend_items: list[str] | None = None
) -> list[str]:
    """Wrap Mermaid markdown and an optional HTML legend in one visual card."""
    lines = ['<div class="uml-diagram-card" markdown="1">', '']
    lines.extend(diagram_lines)

    if legend_items:
        lines.extend(
            [
                '',
                '<p class="uml-legend__title">Legend</p>',
                '<div class="uml-legend" role="list" aria-label="Diagram relationship legend">',
            ]
        )
        lines.extend(legend_items)
        lines.append('</div>')

    lines.extend(['', '</div>'])
    return lines



from nomad.metainfo.data_type import normalize_type

from nomad_simulations.schema_packages.data_types import Bound, m_float_bounded


def standalone_type_roundtrip() -> tuple[str, str]:
    original = m_float_bounded(dtype=float, bound=Bound('[0,1]'))
    serialized = original.serialize_self()
    reconstructed = normalize_type(serialized)
    return str(original.__class__.__name__), str(reconstructed.__class__.__name__)



import pytest
from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.variables import Variables

from . import logger


class TestVariables:
    """
    Test the `Variables` class defined in `variables.py`.
    """

    @pytest.mark.parametrize(
        'n_points, points, result',
        [
            (3, [-1, 0, 1], 3),
            (5, [-1, 0, 1], 3),
            (None, [-1, 0, 1], 3),
            (4, None, 4),
            (4, [], 4),
        ],
    )
    def test_normalize(self, n_points: int, points: list, result: int):
        """
        Test the `normalize` and `get_n_points` methods.
        """
        variable = Variables(
            name='variable_1',
            n_points=n_points,
            points=points,
        )
        assert variable.get_n_points(logger=logger) == result
        variable.normalize(EntryArchive(), logger)
        assert variable.n_points == result

    def test_name_setting_during_normalization(self):
        """
        Test that the name is set during normalization for Variables.
        """
        variable = Variables()
        variable.normalize(EntryArchive(), logger)
        assert variable.name == 'Variables'



import os
from collections.abc import Callable
from typing import Any

import numpy as np
import pytest
import structlog
from nomad.datamodel import EntryArchive
from nomad.units import ureg
from structlog.testing import LogCapture

from nomad_simulations.schema_packages.atoms_state import (
    AtomsState,
    BaseSpinOrbitalState,
    ElectronicState,
    SphericalSymmetryState,
)
from nomad_simulations.schema_packages.general import Simulation
from nomad_simulations.schema_packages.model_method import ModelMethod
from nomad_simulations.schema_packages.model_system import ModelSystem, Representation
from nomad_simulations.schema_packages.numerical_settings import (
    KLinePath as KLinePathSettings,
)
from nomad_simulations.schema_packages.numerical_settings import (
    KMesh as KMeshSettings,
)
from nomad_simulations.schema_packages.numerical_settings import (
    KSpace,
)
from nomad_simulations.schema_packages.outputs import (
    Outputs,
)
from nomad_simulations.schema_packages.properties import (
    DOSProfile,
    ElectronicBandGap,
    ElectronicBandStructure,
    ElectronicDensityOfStates,
)
from nomad_simulations.schema_packages.variables import Energy2 as Energy
from nomad_simulations.schema_packages.variables import KLinePath

from . import logger

# Backwards-compatible alias for tests still using AtomicCell typing.
AtomicCell = Representation

if os.getenv('_PYTEST_RAISE', '0') != '0':

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


def generate_simulation(
    model_system: list[ModelSystem] = [],
    model_method: list[ModelMethod] = [],
    outputs: list[Outputs] = [],
) -> Simulation:
    """
    Generate a `Simulation` section with the main sub-sections, `ModelSystem`, `ModelMethod`, and `Outputs`.
    """
    return Simulation(
        model_method=model_method,
        model_system=model_system,
        outputs=outputs,
    )


def generate_model_system(
    system_type: str | None = None,
    positions: list[list[float]] | None = None,
    lattice_vectors: list[list[float]] | None = None,
    chemical_symbols: list[str] | None = None,
    orbitals_symbols: list[list[str]] | None = None,
    is_representative: bool | None = None,
    pbc: list[bool] | None = None,
    representation_name: str | None = None,
) -> ModelSystem | None:
    """
    Generate a `ModelSystem` section with the given parameters.
    All parameters are optional and will only be set if explicitly provided.
    Default values are not automatically applied.
    """
    if (
        chemical_symbols is not None
        and orbitals_symbols is not None
        and len(chemical_symbols) != len(orbitals_symbols)
    ):
        return None

    # Create model system with only provided values
    model_system_kwargs: dict[str, Any] = {}
    if system_type is not None:
        model_system_kwargs['type'] = system_type
    if is_representative is not None:
        model_system_kwargs['is_representative'] = is_representative

    model_system = ModelSystem(**model_system_kwargs)

    # Set positions if provided
    if positions is not None:
        model_system.positions = np.array(positions) * ureg.angstrom

    # Create representation with only provided values
    representation_kwargs: dict = {}
    if lattice_vectors is not None:
        representation_kwargs['lattice_vectors'] = (
            np.array(lattice_vectors) * ureg.angstrom
        )
    if pbc is not None and all(isinstance(x, bool) for x in pbc):
        representation_kwargs['periodic_boundary_conditions'] = pbc
    if representation_name is not None:
        representation_kwargs['name'] = representation_name

    # Only append representation if we have at least one parameter
    if representation_kwargs:
        model_system.representations.append(Representation(**representation_kwargs))

    # Add atoms_state to the model_system using ElectronicState with basis_orbitals
    if chemical_symbols is not None and orbitals_symbols is not None:
        atoms_state = []
        for element, orbitals in zip(chemical_symbols, orbitals_symbols):
            basis_list: list[SphericalSymmetryState] = []
            for o in orbitals:
                # Map common orbital labels like 's', 'px', 'py', 'pz', 'xy', 'xz', 'z^2', 'yz', 'x^2-y^2'
                l_symbol_to_number = {'s': 0, 'p': 1, 'd': 2, 'f': 3}
                ml_p_symbols = {'x': -1, 'z': 0, 'y': 1}
                ml_d_symbols = {'xy': -2, 'xz': -1, 'z^2': 0, 'yz': 1, 'x^2-y^2': 2}

                if o in {'s', 'p', 'd', 'f'}:
                    state = SphericalSymmetryState()
                    state.l_quantum_number = l_symbol_to_number[o]
                else:
                    # split into l-symbol and ml-symbol when possible
                    # p-orbitals: px, py, pz
                    if o in {'px', 'py', 'pz'}:
                        state = SphericalSymmetryState()
                        state.l_quantum_number = 1  # p
                        state.ml_quantum_number = ml_p_symbols[o[-1]]
                    elif o in {'xy', 'xz', 'z^2', 'yz', 'x^2-y^2'}:
                        state = SphericalSymmetryState()
                        state.l_quantum_number = 2  # d
                        state.ml_quantum_number = ml_d_symbols[o]
                    else:
                        # Fallback: default to s
                        state = SphericalSymmetryState()
                        state.l_quantum_number = 0  # s
                basis_list.append(state)

            e_state = ElectronicState()
            for bo in basis_list:
                e_state.basis_orbitals.append(bo)

            atom_state = AtomsState(chemical_symbol=element)
            atom_state.electronic_state = e_state
            atom_state.normalize(EntryArchive(), logger)
            atoms_state.append(atom_state)

        for state in atoms_state:
            model_system.particle_states.append(state)
    return model_system


def generate_atomic_cell(
    lattice_vectors: list[list[float]] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
    periodic_boundary_conditions: list[bool] = [False, False, False],
    chemical_symbols: list[str] = ['H', 'H', 'O'],
    atomic_numbers: list[int] = [1, 1, 8],
) -> AtomicCell:
    """
    Generate an `AtomicCell` section with the given parameters.
    """

    # Define the atomic cell solely with cell properties; positions are handled by ModelSystem.
    atomic_cell = AtomicCell(periodic_boundary_conditions=periodic_boundary_conditions)
    if lattice_vectors:
        atomic_cell.lattice_vectors = np.array(lattice_vectors) * ureg('angstrom')
    # Removed assignment of positions to atomic_cell as of nomad-simulations>=0.4.
    # Also, the atoms_state information is now part of ModelSystem (particle_states) instead.

    return atomic_cell


def generate_simulation_electronic_dos(
    energy_points: list[int] = [-3, -2, -1, 0, 1, 2, 3],
) -> Simulation:
    """
    Generate a `Simulation` section with an `ElectronicDensityOfStates` section under `Outputs`. It uses
    the template of the model_system created with the `generate_model_system` function.
    """
    # Create the `Simulation` section to make refs work
    model_system = generate_model_system(
        representation_name='original',
        system_type='bulk',
        positions=[[0, 0, 0], [0.5, 0.5, 0.5]],
        lattice_vectors=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        chemical_symbols=['Ga', 'As'],
        orbitals_symbols=[['s'], ['px', 'py']],
        is_representative=True,
        pbc=[False, False, False],
    )
    outputs = Outputs()
    simulation = generate_simulation(model_system=[model_system], outputs=[outputs])

    outputs.normalize(EntryArchive(), logger)

    # Populating the `ElectronicDensityOfStates` section
    variables_energy = Energy(points=energy_points * ureg.joule)
    electronic_dos = ElectronicDensityOfStates(energies=variables_energy)
    outputs.electronic_dos.append(electronic_dos)
    # Build ElectronicState sub_states for specific orbitals and add them to AtomsState
    ga_atom = model_system.particle_states[0]
    as_atom = model_system.particle_states[1]

    # Create sub_states for Ga atom orbitals (s)
    ga_s_basis = ga_atom.electronic_state.basis_orbitals[0]
    es_ga_s = ElectronicState(spin_orbit_state=ga_s_basis)
    ga_atom.electronic_state.sub_states.append(es_ga_s)
    es_ga_s.normalize(EntryArchive(), logger)

    # Create sub_states for As atom orbitals (px, py)
    as_px_basis = as_atom.electronic_state.basis_orbitals[0]
    es_as_px = ElectronicState(spin_orbit_state=as_px_basis)
    as_atom.electronic_state.sub_states.append(es_as_px)
    es_as_px.normalize(EntryArchive(), logger)

    as_py_basis = as_atom.electronic_state.basis_orbitals[1]
    es_as_py = ElectronicState(spin_orbit_state=as_py_basis)
    as_atom.electronic_state.sub_states.append(es_as_py)
    es_as_py.normalize(EntryArchive(), logger)

    # Normalize the top-level ElectronicState to set their names
    ga_atom.electronic_state.normalize(EntryArchive(), logger)
    as_atom.electronic_state.normalize(EntryArchive(), logger)

    orbital_s_Ga_pdos = DOSProfile(
        energies=variables_energy,
        entity_ref=es_ga_s,
    )
    orbital_px_As_pdos = DOSProfile(
        energies=variables_energy,
        entity_ref=es_as_px,
    )
    orbital_py_As_pdos = DOSProfile(
        energies=variables_energy,
        entity_ref=es_as_py,
    )
    orbital_s_Ga_pdos.value = [0.2, 0.5, 0, 0, 0, 0.0, 0.0] * ureg('1/joule')
    orbital_px_As_pdos.value = [1.0, 0.2, 0, 0, 0, 0.3, 0.0] * ureg('1/joule')
    orbital_py_As_pdos.value = [0.3, 0.5, 0, 0, 0, 0.5, 1.3] * ureg('1/joule')
    electronic_dos.projected_dos = [
        orbital_s_Ga_pdos,
        orbital_px_As_pdos,
        orbital_py_As_pdos,
    ]
    return simulation


def generate_k_line_path(
    high_symmetry_path_names: list[str] = ['Gamma', 'X', 'Y', 'Gamma'],
    high_symmetry_path_values: list[list[float]] = [
        [0, 0, 0],
        [0.5, 0, 0],
        [0, 0.5, 0],
        [0, 0, 0],
    ],
) -> KLinePathSettings:
    return KLinePathSettings(
        high_symmetry_path_names=high_symmetry_path_names,
        high_symmetry_path_values=high_symmetry_path_values,
    )


def generate_k_space_simulation(
    system_type: str = 'bulk',
    is_representative: bool = True,
    positions: list[list[float]] = [[0, 0, 0], [0.5, 0.5, 0.5]],
    lattice_vectors: list[list[float]] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
    chemical_symbols: list[str] = ['Ga', 'As'],
    orbitals_symbols: list[list[str]] = [['s'], ['px', 'py']],
    pbc: list[bool] = [False, False, False],
    reciprocal_lattice_vectors: list[list[float]] | None = [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
    ],
    high_symmetry_path_names: list[str] = ['Gamma', 'X', 'Y', 'Gamma'],
    high_symmetry_path_values: list[list[float]] = [
        [0, 0, 0],
        [0.5, 0, 0],
        [0, 0.5, 0],
        [0, 0, 0],
    ],
    klinepath_points: list[float] | None = None,
    grid=[6, 6, 6],
) -> Simulation:
    model_system = generate_model_system(
        representation_name='primitive',
        system_type=system_type,
        is_representative=is_representative,
        positions=positions,
        lattice_vectors=lattice_vectors,
        chemical_symbols=chemical_symbols,
        orbitals_symbols=orbitals_symbols,
        pbc=pbc,
    )
    k_space = KSpace()
    # adding `reciprocal_lattice_vectors`
    if reciprocal_lattice_vectors is not None:
        k_space.reciprocal_lattice_vectors = (
            2 * np.pi * np.array(reciprocal_lattice_vectors) / ureg.angstrom
        )
    # adding `KMeshSettings
    k_mesh = KMeshSettings(grid=grid)
    k_space.k_mesh.append(k_mesh)
    # adding `KLinePathSettings`
    k_line_path = KLinePathSettings(
        high_symmetry_path_names=high_symmetry_path_names,
        high_symmetry_path_values=high_symmetry_path_values,
    )
    if klinepath_points is not None:
        k_line_path.points = klinepath_points
    k_space.k_line_path = k_line_path
    # appending `KSpace` to `ModelMethod.numerical_settings`
    model_method = ModelMethod()
    model_method.numerical_settings.append(k_space)
    return generate_simulation(model_method=[model_method], model_system=[model_system])


def generate_electronic_band_structure(
    reciprocal_lattice_vectors: list[list[float]] | None = [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
    ],
    value: list | None = [
        [3, -2],
        [3, 1],
        [4, -2],
        [5, -1],
        [4, 0],
        [2, 0],
        [2, 1],
        [4, -3],
    ],
    occupation: list | None = [
        [0, 2],
        [0, 1],
        [0, 2],
        [0, 2],
        [0, 1.5],
        [0, 1.5],
        [0, 1],
        [0, 2],
    ],
    highest_occupied: float | None = None,
    lowest_unoccupied: float | None = None,
) -> ElectronicBandStructure:
    """
    Generate an `ElectronicBandStructure` section with the given parameters.
    """
    outputs = Outputs()
    k_space = KSpace(
        k_line_path=KLinePathSettings(
            points=[
                [0, 0, 0],
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [1, 1, 0],
                [1, 0, 1],
                [0, 1, 1],
                [1, 1, 1],
            ]
        )
    )
    model_method = ModelMethod(numerical_settings=[k_space])
    if reciprocal_lattice_vectors:
        k_space.reciprocal_lattice_vectors = reciprocal_lattice_vectors
    _ = generate_simulation(
        model_system=[
            generate_model_system(
                representation_name='original',
                system_type='bulk',
                positions=[[0, 0, 0], [0.5, 0.5, 0.5]],
                lattice_vectors=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                chemical_symbols=['Ga', 'As'],
                orbitals_symbols=[['s'], ['px', 'py']],
                is_representative=True,
                pbc=[False, False, False],
            )
        ],
        model_method=[model_method],
        outputs=[outputs],
    )
    electronic_band_structure = ElectronicBandStructure(n_levels=2)
    outputs.electronic_eigenvalues = [electronic_band_structure]
    electronic_band_structure.k_path = KLinePath(
        points=model_method.numerical_settings[0].k_line_path
    )
    if value is not None:
        electronic_band_structure.value = value
    if occupation is not None:
        electronic_band_structure.occupation = occupation
    electronic_band_structure.highest_occupied = highest_occupied
    electronic_band_structure.lowest_unoccupied = lowest_unoccupied
    return electronic_band_structure


@pytest.fixture(scope='session')
def model_system() -> ModelSystem:
    return generate_model_system(
        representation_name='original',
        system_type='bulk',
        positions=[[0, 0, 0], [0.5, 0.5, 0.5]],
        lattice_vectors=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        chemical_symbols=['Ga', 'As'],
        orbitals_symbols=[['s'], ['px', 'py']],
        is_representative=True,
        pbc=[False, False, False],
    )


@pytest.fixture(scope='session')
def simulation_electronic_dos() -> Simulation:
    return generate_simulation_electronic_dos()


@pytest.fixture(scope='session')
def k_line_path() -> KLinePathSettings:
    return generate_k_line_path()


@pytest.fixture(scope='session')
def k_space_simulation() -> Simulation:
    return generate_k_space_simulation()


refs_apw = [
    {
        'm_def': 'nomad_simulations.schema_packages.basis_set.BasisSetContainer',
    },
    {
        'm_def': 'nomad_simulations.schema_packages.basis_set.BasisSetContainer',
        'basis_set_components': [
            {
                'm_def': 'nomad_simulations.schema_packages.basis_set.APWPlaneWaveBasisSet',
                'cutoff_energy': 500.0,
            },
        ],
    },
    {
        'm_def': 'nomad_simulations.schema_packages.basis_set.BasisSetContainer',
        'basis_set_components': [
            {
                'm_def': 'nomad_simulations.schema_packages.basis_set.APWPlaneWaveBasisSet',
                'cutoff_energy': 500.0,
            },
            {
                'm_def': 'nomad_simulations.schema_packages.basis_set.MuffinTinRegion',
                'species_scope': ['/data/model_system/0/particle_states/0'],
                'radius': 1.0,
                'l_max': 2,
                'l_channels': [
                    {
                        'name': 0,
                        'orbitals': [
                            {
                                'm_def': 'nomad_simulations.schema_packages.basis_set.APWOrbital',
                                'energy_parameter': [0.0],
                                'differential_order': [0],
                            },
                        ],
                    },
                    {
                        'name': 1,
                        'orbitals': [
                            {
                                'm_def': 'nomad_simulations.schema_packages.basis_set.APWOrbital',
                                'energy_parameter': [0.0],
                                'differential_order': [0],
                            },
                        ],
                    },
                    {
                        'name': 2,
                        'orbitals': [
                            {
                                'm_def': 'nomad_simulations.schema_packages.basis_set.APWOrbital',
                                'energy_parameter': [0.0],
                                'differential_order': [0],
                            },
                        ],
                    },
                ],
            },
        ],
    },
]


@pytest.fixture
def log_output():
    capture = LogCapture()
    processors = structlog.get_config()['processors']
    old_processors = processors.copy()
    try:
        # clear processors list and use LogCapture for testing
        processors.clear()
        processors.append(capture)
        structlog.configure(processors=processors)
        yield capture
    finally:
        # remove LogCapture and restore original processors
        processors.clear()
        processors.extend(old_processors)
        structlog.configure(processors=processors)


@pytest.fixture
def approx():
    def func(expected, abs: float = 0.0, rel=1e-6):
        return pytest.approx(expected, abs=abs, rel=rel)

    return func


def make_spherical_state(
    n: int | None = 1,
    lqn: int | None = 0,
    ml: int | None = None,
    j: float | None = None,
    kappa: int | None = None,
    s: float | None = None,
    ms: float | None = None,
    coupling_origin: str | None = None,
) -> SphericalSymmetryState:
    """
    Construct a SphericalSymmetryState with convenient aliases for quantum numbers.

    Parameters map to schema quantities:
    - n -> n_quantum_number
    - lqn -> l_quantum_number
    - ml -> ml_quantum_number
    - j -> j_quantum_number (float)
    - kappa -> kappa_quantum_number
    - s -> s_quantum_number
    - ms -> ms_quantum_number
    - coupling_origin -> coupling_origin (e.g., 'pure_LS','pure_jj','intermediate','relativistic')
    """
    state = SphericalSymmetryState()
    if n is not None:
        state.n_quantum_number = n
    if lqn is not None:
        state.l_quantum_number = lqn
    if ml is not None:
        state.ml_quantum_number = ml
    if j is not None:
        state.j_quantum_number = j
    if kappa is not None:
        state.kappa_quantum_number = kappa
    if s is not None:
        state.s_quantum_number = s
    if ms is not None:
        state.ms_quantum_number = ms
    if coupling_origin is not None:
        state.coupling_origin = coupling_origin
    return state


def make_electronic_state(
    basis_orbitals: list[BaseSpinOrbitalState] | None = None,
    *,
    name: str | None = None,
    symmetry_label: str | None = None,
    point_group: str | None = None,
    occupation: float | None = None,
    spin_orbit: BaseSpinOrbitalState | None = None,
    sub_states: list[ElectronicState] | None = None,
) -> ElectronicState:
    """
    Construct an ElectronicState, optionally with a list of basis_orbitals and other metadata.

    - basis_orbitals: list of BaseSpinOrbitalState (e.g., SphericalSymmetryState) to define expansion basis
    - name, symmetry_label, point_group, occupation: metadata
    - spin_orbit: optional BaseSpinOrbitalState to assign to spin_orbit_state subsection
    - sub_states: optional child ElectronicState hierarchy to append to sub_states
    """
    es = ElectronicState()
    if name is not None:
        es.name = name
    if symmetry_label is not None:
        es.symmetry_label = symmetry_label
    if point_group is not None:
        es.point_group = point_group
    if occupation is not None:
        es.occupation = occupation

    if spin_orbit is not None:
        es.spin_orbit_state = spin_orbit

    if basis_orbitals:
        for bo in basis_orbitals:
            es.basis_orbitals.append(bo)

    if sub_states:
        for child in sub_states:
            es.sub_states.append(child)

    return es



"""
Tests for the convergence target classes.

This module tests the refactored class-based convergence target system including:
- Base class helper methods for convergence checking
- Individual convergence target classes (Energy, Force, Potential, Charge)
- Convergence type enum validation
- Unit handling and is_reached flag calculation
"""

import numpy as np
import pytest
from nomad.units import ureg

from nomad_simulations.schema_packages.outputs import Outputs, SCFSteps
from nomad_simulations.schema_packages.properties import TotalEnergy, TotalForce
from nomad_simulations.schema_packages.workflow.general import (
    ChargeConvergenceTarget,
    EnergyConvergenceTarget,
    ForceConvergenceTarget,
    PotentialConvergenceTarget,
    SimulationWorkflow,
    SimulationWorkflowMethod,
    WavefunctionConvergenceTarget,
)


class TestEnergyConvergenceTarget:
    """Test the EnergyConvergenceTarget class."""

    @pytest.mark.parametrize(
        'threshold, threshold_type, energy_values, expected_reached',
        [
            # Absolute convergence - converged
            (1e-6 * ureg.joule, 'absolute', [1e-10, 1e-11, 5e-12], True),
            # Absolute convergence - not converged
            (1e-6 * ureg.joule, 'absolute', [1e-3, 1e-4, 1e-5], False),
        ],
    )
    def test_energy_convergence(
        self,
        threshold: float,
        threshold_type: str,
        energy_values: list[float],
        expected_reached: bool,
        archive,
        logger,
    ):
        """
        Test energy convergence checking with different thresholds and types.

        Args:
            threshold: Convergence threshold value.
            threshold_type: Type of convergence check ('absolute' or 'relative').
            energy_values: List of energy values to test.
            expected_reached: Expected value of is_reached flag.
        """
        # Set up convergence target
        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = threshold
        energy_target.threshold_type = threshold_type

        # Create SCF steps with energy delta values
        scf_step = SCFSteps()
        if len(energy_values) > 1:
            # delta_energies_total is an array of changes at each SCF step
            deltas = [
                abs(energy_values[i] - energy_values[i - 1])
                for i in range(1, len(energy_values))
            ]
            scf_step.delta_energies_total = np.array(deltas) * ureg.joule

        # Add outputs to archive
        archive.data.outputs = [Outputs(scf_steps=scf_step)]

        # Normalize and check
        is_reached = energy_target.normalize(archive, logger)

        if len(energy_values) > 1 and energy_values[-1] != 0:
            assert is_reached == expected_reached

    def test_energy_units(self, archive, logger):
        """Test that energy convergence handles units correctly."""
        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = 1e-6 * ureg.joule
        energy_target.threshold_type = 'absolute'

        # Create SCF steps with energy in different unit (hartree)
        scf_step = SCFSteps()
        scf_step.delta_energies_total = (
            np.array([1e-7]) * ureg.hartree
        )  # Should convert to joule

        archive.data.outputs = [Outputs(scf_steps=scf_step)]
        is_reached = energy_target.normalize(archive, logger)

        # Check that conversion happened and convergence was checked
        assert is_reached is not None


class TestForceConvergenceTarget:
    """Test the ForceConvergenceTarget class."""

    @pytest.mark.parametrize(
        'threshold, threshold_type, force_values, expected_reached',
        [
            # Maximum convergence - converged
            (
                1e-8 * ureg.newton,
                'maximum',
                np.array([[1e-10, 2e-10, 3e-10], [1e-11, 1e-11, 1e-11]]),
                True,
            ),
            # Maximum convergence - not converged
            (
                1e-8 * ureg.newton,
                'maximum',
                np.array([[1e-5, 2e-6, 3e-6], [1e-6, 1e-7, 1e-7]]),
                False,
            ),
            # RMS convergence - converged
            (
                1e-8 * ureg.newton,
                'rms',
                np.array([[1e-10, 1e-10, 1e-10], [1e-10, 1e-10, 1e-10]]),
                True,
            ),
            # RMS convergence - not converged
            (
                1e-8 * ureg.newton,
                'rms',
                np.array([[1e-5, 1e-5, 1e-5], [1e-6, 1e-6, 1e-6]]),
                False,
            ),
            # Single atom case
            (1e-8 * ureg.newton, 'maximum', np.array([[1e-10, 1e-10, 1e-10]]), True),
        ],
    )
    def test_force_convergence(
        self,
        threshold: float,
        threshold_type: str,
        force_values: np.ndarray,
        expected_reached: bool,
        archive,
        logger,
    ):
        """
        Test force convergence with maximum and RMS types.

        Args:
            threshold: Convergence threshold value in newton.
            threshold_type: Type of convergence check ('maximum' or 'rms').
            force_values: Array of force values (n_atoms, 3).
            expected_reached: Expected value of is_reached flag.
        """
        force_target = ForceConvergenceTarget()
        force_target.threshold = threshold
        force_target.threshold_type = threshold_type

        # Create outputs with forces and scf_steps
        forces = TotalForce(value=force_values * ureg.newton)
        scf_steps = SCFSteps()  # Empty scf_steps for normalization to populate
        outputs = Outputs(total_forces=[forces], scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Normalize outputs to compute delta_force_abs from total_forces
        outputs.normalize(archive, logger)

        # Check convergence
        is_reached = force_target.normalize(archive, logger)
        assert is_reached == expected_reached

    def test_force_absolute_convergence(self, archive, logger):
        """Test absolute force convergence from SCF delta."""
        force_target = ForceConvergenceTarget()
        force_target.threshold = 1e-8 * ureg.newton
        force_target.threshold_type = 'absolute'

        # Create SCF steps with delta_force_abs showing convergence
        scf_step = SCFSteps()
        scf_step.delta_force_abs = np.array([1e-5, 1e-10]) * ureg.newton

        archive.data.outputs = [Outputs(scf_steps=scf_step)]

        is_reached = force_target.normalize(archive, logger)
        # Last delta force (1e-10) should be below threshold (1e-8)
        assert is_reached is True


class TestPotentialConvergenceTarget:
    """Test the PotentialConvergenceTarget class."""

    @pytest.mark.parametrize(
        'threshold, threshold_type, potential_values, expected_reached',
        [
            # RMS convergence - converged
            (1e-5 * ureg.joule, 'rms', np.array([1e-7, 1e-7, 1e-7, 1e-7]), True),
            # RMS convergence - not converged
            (1e-5 * ureg.joule, 'rms', np.array([1e-3, 1e-3, 1e-3, 1e-3]), False),
            # Absolute convergence - converged
            (1e-5 * ureg.joule, 'absolute', np.array([1e-8]), True),
            # Absolute convergence - not converged
            (1e-5 * ureg.joule, 'absolute', np.array([1e-3]), False),
        ],
    )
    def test_potential_convergence(
        self,
        threshold: float,
        threshold_type: str,
        potential_values: np.ndarray,
        expected_reached: bool,
        archive,
        logger,
    ):
        """
        Test potential convergence with RMS and absolute types.

        Args:
            threshold: Convergence threshold value in joule.
            threshold_type: Type of convergence check ('rms' or 'absolute').
            potential_values: Array of potential values.
            expected_reached: Expected value of is_reached flag.
        """
        potential_target = PotentialConvergenceTarget()
        potential_target.threshold = threshold
        potential_target.threshold_type = threshold_type

        # Create SCF steps with potential RMS values
        scf_step = SCFSteps()
        scf_step.delta_potential_rms = potential_values * ureg.joule

        archive.data.outputs = [Outputs(scf_steps=scf_step)]

        # Normalize and check
        is_reached = potential_target.normalize(archive, logger)
        assert is_reached == expected_reached


class TestChargeConvergenceTarget:
    """Test the ChargeConvergenceTarget class."""

    @pytest.mark.parametrize(
        'threshold, threshold_type, charge_values, expected_reached',
        [
            # Absolute convergence - converged
            (1e-7 * ureg.coulomb, 'absolute', np.array([1e-10, 1e-10, 1e-10]), True),
            # Absolute convergence - not converged
            (1e-7 * ureg.coulomb, 'absolute', np.array([1e-5, 1e-5, 1e-5]), False),
            # RMS convergence - converged
            (1e-7 * ureg.coulomb, 'rms', np.array([1e-10, 1e-10, 1e-10, 1e-10]), True),
            # RMS convergence - not converged
            (1e-7 * ureg.coulomb, 'rms', np.array([1e-5, 1e-5, 1e-5, 1e-5]), False),
        ],
    )
    def test_charge_convergence(
        self,
        threshold: float,
        threshold_type: str,
        charge_values: np.ndarray,
        expected_reached: bool,
        archive,
        logger,
    ):
        """
        Test charge convergence with absolute and RMS types.

        Args:
            threshold: Convergence threshold (dimensionless).
            threshold_type: Type of convergence check ('absolute' or 'rms').
            charge_values: Array of charge difference values.
            expected_reached: Expected value of is_reached flag.
        """
        charge_target = ChargeConvergenceTarget()
        charge_target.threshold = threshold
        charge_target.threshold_type = threshold_type

        # Create SCF steps with density RMS values (charge convergence)
        scf_step = SCFSteps()
        scf_step.delta_density_rms = charge_values * ureg.coulomb

        archive.data.outputs = [Outputs(scf_steps=scf_step)]

        # Normalize and check
        is_reached = charge_target.normalize(archive, logger)
        assert is_reached == expected_reached


class TestWavefunctionConvergenceTarget:
    """Test the WavefunctionConvergenceTarget class."""

    @pytest.mark.parametrize(
        'threshold, threshold_type, wf_values, expected_reached',
        [
            # Absolute convergence - converged
            (1e-8 * ureg.dimensionless, 'absolute', [1e-10, 1e-11, 5e-12], True),
            # Absolute convergence - not converged
            (1e-8 * ureg.dimensionless, 'absolute', [1e-5, 1e-6, 1e-7], False),
            # RMS convergence with array data - converged
            (1e-8 * ureg.dimensionless, 'rms', [[1e-10, 2e-10], [5e-11, 6e-11]], True),
            # RMS convergence with array data - not converged
            (1e-8 * ureg.dimensionless, 'rms', [[1e-5, 2e-5], [8e-6, 9e-6]], False),
            # Maximum convergence - not converged
            (
                1e-8 * ureg.dimensionless,
                'maximum',
                [[1e-7, 1e-10], [8e-7, 2e-10]],
                False,
            ),
            # Values right at threshold boundary - not converged
            (1e-8 * ureg.dimensionless, 'absolute', [0.0, 1.1e-8, 2.2e-8], False),
            # Very small values near machine epsilon
            (1e-16 * ureg.dimensionless, 'absolute', [1e-17, 5e-18, 1e-18], True),
        ],
    )
    def test_wavefunction_convergence(
        self,
        threshold: float,
        threshold_type: str,
        wf_values: list,
        expected_reached: bool,
        archive,
        logger,
    ):
        """
        Test wavefunction convergence with different thresholds and data types.

        Args:
            threshold: Convergence threshold (dimensionless).
            threshold_type: Type of convergence check.
            wf_values: List of wavefunction values (scalar or array).
            expected_reached: Expected value of is_reached flag.
        """
        wavefunction_target = WavefunctionConvergenceTarget()
        wavefunction_target.threshold = threshold
        wavefunction_target.threshold_type = threshold_type

        scf_step = SCFSteps()
        if len(wf_values) > 1:
            # Convert to deltas based on data type
            if isinstance(wf_values[0], list):
                # Array data - compute element-wise deltas
                deltas = []
                for i in range(1, len(wf_values)):
                    delta = np.abs(np.array(wf_values[i]) - np.array(wf_values[i - 1]))
                    deltas.append(delta)
                scf_step.delta_wavefunction_rms = deltas
            else:
                # Scalar data - compute simple deltas
                deltas = [
                    abs(wf_values[i] - wf_values[i - 1])
                    for i in range(1, len(wf_values))
                ]
                scf_step.delta_wavefunction_rms = np.array(deltas)

        archive.data.outputs = [Outputs(scf_steps=scf_step)]
        is_reached = wavefunction_target.normalize(archive, logger)
        assert is_reached == expected_reached


class TestMissingDataHandling:
    """Test that all convergence targets handle missing data correctly."""

    @pytest.mark.parametrize(
        'target_class, threshold, threshold_unit',
        [
            (EnergyConvergenceTarget, 1e-6, ureg.joule),
            (ForceConvergenceTarget, 1e-8, ureg.newton),
            (PotentialConvergenceTarget, 1e-5, ureg.joule),
            (ChargeConvergenceTarget, 1e-7, ureg.coulomb),
            (WavefunctionConvergenceTarget, 1e-8, ureg.dimensionless),
        ],
        ids=['Energy', 'Force', 'Potential', 'Charge', 'Wavefunction'],
    )
    def test_missing_data_returns_none(
        self, target_class, threshold, threshold_unit, archive, logger
    ):
        """Test convergence targets return None when data is missing."""
        target = target_class()
        target.threshold = threshold * threshold_unit
        target.threshold_type = 'absolute'

        # Empty archive
        is_reached = target.normalize(archive, logger)
        assert is_reached is None

        # Archive with outputs but no SCF data
        archive.data.outputs = [Outputs()]
        is_reached = target.normalize(archive, logger)
        assert is_reached is None


class TestConvergenceHelperMethods:
    """Test the base class helper methods for convergence checking."""

    def test_check_absolute(self, logger):
        """Test _check_absolute helper method."""
        energy_target = EnergyConvergenceTarget()
        # Set threshold with units
        energy_target.threshold = 1e-6 * ureg.joule

        # Test converged case
        assert energy_target._check_absolute(5e-7 * ureg.joule, logger) is True
        assert energy_target._check_absolute(-5e-7 * ureg.joule, logger) is True

        # Test not converged case
        assert energy_target._check_absolute(5e-5 * ureg.joule, logger) is False
        assert energy_target._check_absolute(-5e-5 * ureg.joule, logger) is False

        # Test boundary (now uses <=)
        assert energy_target._check_absolute(1e-6 * ureg.joule, logger) is True
        assert energy_target._check_absolute(9e-7 * ureg.joule, logger) is True

    def test_check_relative(self, logger):
        """Test _check_relative helper method."""
        energy_target = EnergyConvergenceTarget()
        # For relative convergence, threshold must be dimensionless
        # Use raw float since EnergyConvergenceTarget expects joule unit
        energy_target.threshold = 1e-6

        # Test converged case
        assert (
            energy_target._check_relative(1e-7 * ureg.joule, 1.0 * ureg.joule, logger)
            is True
        )
        assert (
            energy_target._check_relative(-1e-7 * ureg.joule, 1.0 * ureg.joule, logger)
            is True
        )

        # Test not converged case
        assert (
            energy_target._check_relative(1e-3 * ureg.joule, 1.0 * ureg.joule, logger)
            is False
        )

        # Test with different reference values
        assert (
            energy_target._check_relative(1e-6 * ureg.joule, 100.0 * ureg.joule, logger)
            is True
        )
        assert (
            energy_target._check_relative(1e-4 * ureg.joule, 100.0 * ureg.joule, logger)
            is False
        )

        # Test edge case: zero reference
        assert (
            energy_target._check_relative(0.0 * ureg.joule, 0.0 * ureg.joule, logger)
            is True
        )

    def test_check_maximum(self, logger):
        """Test _check_maximum helper method."""
        # Create a force target with units
        force_target = ForceConvergenceTarget()
        force_target.threshold = 1e-8 * ureg.newton

        # Test converged case
        values_converged = np.array([1e-10, -5e-10, 3e-10]) * ureg.newton
        assert force_target._check_maximum(values_converged, logger) is True

        # Test not converged case
        values_not_converged = np.array([1e-5, 1e-6, 1e-7]) * ureg.newton
        assert force_target._check_maximum(values_not_converged, logger) is False

        # Test multidimensional array
        values_2d = np.array([[1e-10, 2e-10], [3e-10, 4e-10]]) * ureg.newton
        assert force_target._check_maximum(values_2d, logger) is True

    def test_check_rms(self, logger):
        """Test _check_rms helper method."""
        # Create a force target with units
        force_target = ForceConvergenceTarget()
        force_target.threshold = 1e-8 * ureg.newton

        # Test converged case
        values_converged = np.array([1e-10, 1e-10, 1e-10]) * ureg.newton
        assert force_target._check_rms(values_converged, logger) is True

        # Test not converged case
        values_not_converged = np.array([1e-5, 1e-5, 1e-5]) * ureg.newton
        assert force_target._check_rms(values_not_converged, logger) is False

        # Test with zeros
        values_zero = np.array([0.0, 0.0, 0.0]) * ureg.newton
        assert force_target._check_rms(values_zero, logger) is True

        # Test multidimensional array
        values_2d = np.array([[1e-10, 2e-10], [3e-10, 4e-10]]) * ureg.newton
        rms = np.sqrt(np.mean(values_2d**2))
        assert force_target._check_rms(values_2d, logger) == (rms < 1e-8 * ureg.newton)


class TestConvergenceTypeEnumeration:
    """Test threshold_type enum validation."""

    @pytest.mark.parametrize(
        'threshold_type, threshold',
        [
            ('absolute', 1e-6 * ureg.joule),
            ('relative', 1e-6),
            ('maximum', 1e-6 * ureg.joule),
            ('rms', 1e-6 * ureg.joule),
        ],
    )
    def test_valid_convergence_types(
        self, threshold_type: str, threshold, archive, logger
    ):
        """Test that all valid convergence types are accepted."""
        energy_target = EnergyConvergenceTarget()
        energy_target.threshold_type = threshold_type
        energy_target.threshold = threshold

        # Should not raise an error
        energy_target.normalize(archive, logger)

    def test_default_convergence_type(self, archive, logger):
        """Test that default threshold_type is handled."""
        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = 1e-6 * ureg.joule
        # Don't set threshold_type explicitly

        # Create test data
        scf_step = SCFSteps()
        scf_step.delta_energies_total = np.array([1e-7]) * ureg.joule

        archive.data.outputs = [Outputs(scf_steps=scf_step)]

        # Should default to 'absolute'
        energy_target.normalize(archive, logger)
        # Should not raise an error


class TestConvergenceInWorkflow:
    """Test convergence targets integration with SimulationWorkflow."""

    def test_multiple_convergence_targets(self, archive, logger):
        """Test workflow with multiple convergence criteria."""
        workflow = SimulationWorkflow()
        workflow.method = SimulationWorkflowMethod()

        # Add multiple convergence targets
        workflow.method.convergence_targets = [
            EnergyConvergenceTarget(
                threshold=1e-6 * ureg.joule, threshold_type='absolute'
            ),
            ForceConvergenceTarget(
                threshold=1e-8 * ureg.newton, threshold_type='maximum'
            ),
            ChargeConvergenceTarget(
                threshold=1e-7 * ureg.coulomb, threshold_type='rms'
            ),
        ]

        # Create test data
        scf_step = SCFSteps()
        scf_step.delta_energies_total = np.array([5e-7]) * ureg.joule
        scf_step.delta_density_rms = np.array([1e-8]) * ureg.coulomb

        forces = TotalForce(
            value=np.array([[1e-9, 1e-9, 1e-9], [1e-9, 1e-9, 1e-9]]) * ureg.newton
        )

        outputs = Outputs(scf_steps=scf_step, total_forces=[forces])
        archive.data.outputs = [outputs]

        # Normalize outputs to compute delta_force_abs
        outputs.normalize(archive, logger)

        # Normalize workflow
        workflow.normalize(archive, logger)

        # Check that convergence targets were normalized
        assert workflow.method.convergence_targets is not None
        assert len(workflow.method.convergence_targets) == 3

        # Normalize individual targets to check convergence status
        convergence_results = []
        for target in workflow.method.convergence_targets:
            is_reached = target.normalize(archive, logger)
            convergence_results.append(is_reached)

        # Check individual convergence
        assert convergence_results[0] is True  # Energy
        assert convergence_results[1] is True  # Force
        assert convergence_results[2] is True  # Charge

    def test_convergence_targets_empty(self, archive, logger):
        """Test workflow with no convergence targets."""
        workflow = SimulationWorkflow()

        # No convergence targets set
        workflow.normalize(archive, logger)

        # Should not raise an error
        assert (
            workflow.method.convergence_targets is None
            or len(workflow.method.convergence_targets) == 0
        )

    def test_mixed_convergence_results(self, archive, logger):
        """Test with some targets converged and others not."""
        workflow = SimulationWorkflow()
        workflow.method = SimulationWorkflowMethod()

        workflow.method.convergence_targets = [
            EnergyConvergenceTarget(
                threshold=1e-6 * ureg.joule, threshold_type='absolute'
            ),  # Will converge
            ForceConvergenceTarget(
                threshold=1e-20 * ureg.newton, threshold_type='maximum'
            ),  # Won't converge
        ]

        # Create test data
        scf_step = SCFSteps()
        scf_step.delta_energies_total = np.array([5e-7]) * ureg.joule  # Converges

        forces = TotalForce(
            value=np.array([[1e-5, 1e-5, 1e-5]]) * ureg.newton
        )  # Doesn't converge

        outputs = Outputs(scf_steps=scf_step, total_forces=[forces])
        archive.data.outputs = [outputs]

        # Normalize outputs to compute delta_force_abs
        outputs.normalize(archive, logger)

        # Normalize targets and capture results
        convergence_results = []
        for target in workflow.method.convergence_targets:
            is_reached = target.normalize(archive, logger)
            convergence_results.append(is_reached)

        assert convergence_results[0] is True
        assert convergence_results[1] is False


class TestFallbackPaths:
    """Test fallback path functionality in convergence targets."""

    def test_force_fallback_to_scf(self, archive, logger):
        """Test ForceConvergenceTarget falls back from workflow to SCF level."""
        force_target = ForceConvergenceTarget()
        force_target.threshold = 1e-8 * ureg.newton
        force_target.threshold_type = 'maximum'

        # Create SCF force data (no workflow2)
        scf_step = SCFSteps()
        scf_step.delta_force_abs = np.array([1e-9, 2e-9, 1.5e-9]) * ureg.newton

        archive.data.outputs = [Outputs(scf_steps=scf_step)]

        # Should use scf_steps.delta_force_abs since workflow2 doesn't exist
        is_reached = force_target.normalize(archive, logger)
        assert is_reached is True  # max([1e-9, 2e-9, 1.5e-9]) < 1e-8

    def test_force_fallback_to_computed(self, archive, logger):
        """Test that delta_force_abs is computed from total_forces during normalization."""
        force_target = ForceConvergenceTarget()
        force_target.threshold = 1e-8 * ureg.newton
        force_target.threshold_type = 'maximum'

        # Only provide total_forces (no workflow2, no delta_force_abs)
        forces = TotalForce(
            value=np.array([[1e-9, 1e-9, 1e-9], [2e-9, 2e-9, 2e-9]]) * ureg.newton
        )
        scf_steps = SCFSteps()  # Empty scf_steps for normalization to populate
        outputs = Outputs(total_forces=[forces], scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Normalize outputs to compute delta_force_abs from total_forces
        outputs.normalize(archive, logger)

        # Verify delta_force_abs was computed
        assert outputs.scf_steps.delta_force_abs is not None

        # Check convergence using computed values
        is_reached = force_target.normalize(archive, logger)
        # max norm ≈ 3.46e-9 < 1e-8
        assert is_reached is True

    def test_single_path_backwards_compatible(self, archive, logger):
        """Test that single 'path' annotation still works."""
        # EnergyConvergenceTarget uses single 'path' annotation
        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = 1e-6 * ureg.joule
        energy_target.threshold_type = 'absolute'

        scf_step = SCFSteps()
        scf_step.delta_energies_total = np.array([5e-7]) * ureg.joule

        archive.data.outputs = [Outputs(scf_steps=scf_step)]

        is_reached = energy_target.normalize(archive, logger)
        assert is_reached is True

    def test_energy_delta_computed_from_total_energies(self, archive, logger):
        """Test that delta_energies_total is computed from total_energies during normalization."""
        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = 1e-6 * ureg.joule
        energy_target.threshold_type = 'absolute'

        # Provide total_energies but not delta_energies_total
        energies = [
            TotalEnergy(value=1.0 * ureg.joule),
            TotalEnergy(value=1.0000001 * ureg.joule),  # Delta: 1e-7
            TotalEnergy(value=1.0000002 * ureg.joule),  # Delta: 1e-7
        ]
        scf_steps = SCFSteps()  # Empty scf_steps for normalization to populate
        outputs = Outputs(total_energies=energies, scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Normalize outputs to compute delta_energies_total
        outputs.normalize(archive, logger)

        # Verify delta_energies_total was computed
        assert outputs.scf_steps.delta_energies_total is not None
        assert len(outputs.scf_steps.delta_energies_total) == 2  # n-1 deltas

        # Check convergence using computed values
        is_reached = energy_target.normalize(archive, logger)
        # Last delta is 1e-7 < 1e-6
        assert is_reached is True


class TestThresholdTypeValidation:
    """Test threshold_type validation, especially for relative convergence."""

    def test_relative_with_dimensionless_value(self, archive, logger, caplog):
        """Test that relative convergence validation passes with dimensionless threshold."""
        import logging

        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = 1e-6  # Dimensionless float
        energy_target.threshold_type = 'relative'

        scf_steps = SCFSteps(delta_energies_total=[1e-7, 5e-8] * ureg.joule)
        outputs = Outputs(scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Validation should pass (dimensionless threshold is valid for relative)
        # But relative convergence implementation is incomplete (warns and returns None)
        with caplog.at_level(logging.WARNING):
            is_reached = energy_target.normalize(archive, logger)

        # Validation passes (no error about dimensionless requirement)
        assert (
            'Relative convergence requires dimensionless threshold' not in caplog.text
        )
        # But implementation returns None with warning
        assert is_reached is None
        assert 'Relative convergence requires reference value' in caplog.text

    def test_relative_with_pint_dimensionless(self, archive, logger, caplog):
        """Test that relative convergence validation passes with Pint dimensionless Quantity."""
        import logging

        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = 1e-6 * ureg.dimensionless
        energy_target.threshold_type = 'relative'

        scf_steps = SCFSteps(delta_energies_total=[1e-7, 5e-8] * ureg.joule)
        outputs = Outputs(scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Validation should pass (dimensionless Pint Quantity is valid for relative)
        # But relative convergence implementation is incomplete (warns and returns None)
        with caplog.at_level(logging.WARNING):
            is_reached = energy_target.normalize(archive, logger)

        # Validation passes (no error about dimensionless requirement)
        assert (
            'Relative convergence requires dimensionless threshold' not in caplog.text
        )
        # But implementation returns None with warning
        assert is_reached is None
        assert 'Relative convergence requires reference value' in caplog.text

    def test_relative_with_units_fails(self, archive, logger, caplog):
        """Test that relative convergence fails validation with dimensional units."""
        import logging

        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = (
            1e-6 * ureg.joule
        )  # Has energy units - invalid for relative
        energy_target.threshold_type = 'relative'

        scf_steps = SCFSteps(delta_energies_total=[1e-7, 5e-8] * ureg.joule)
        outputs = Outputs(scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Should fail validation and return None
        with caplog.at_level(logging.ERROR):
            is_reached = energy_target.normalize(archive, logger)

        assert is_reached is None
        # Should log error with structured data about unit mismatch
        assert '"threshold_type": "relative"' in caplog.text
        assert '"requires_physical_units": false' in caplog.text
        assert '"has_physical_units": true' in caplog.text
        # Should NOT reach the "requires reference" warning (validation fails first)
        assert 'Relative convergence requires reference value' not in caplog.text

    def test_absolute_with_units_succeeds(self, archive, logger):
        """Test that absolute convergence works fine with dimensional units."""
        energy_target = EnergyConvergenceTarget()
        energy_target.threshold = 1e-6 * ureg.joule  # Has units - valid for absolute
        energy_target.threshold_type = 'absolute'

        scf_steps = SCFSteps(delta_energies_total=[1e-7, 5e-8] * ureg.joule)
        outputs = Outputs(scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Should validate and check convergence
        is_reached = energy_target.normalize(archive, logger)
        assert is_reached is True

    def test_maximum_with_units_succeeds(self, archive, logger):
        """Test that maximum convergence works with dimensional units."""
        force_target = ForceConvergenceTarget()
        force_target.threshold = 1e-8 * ureg.newton  # Has units - valid for maximum
        force_target.threshold_type = 'maximum'

        scf_steps = SCFSteps(
            delta_force_abs=[1e-9, 5e-10, 2e-10] * ureg.newton  # Array for maximum
        )
        outputs = Outputs(scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Should validate and check convergence
        is_reached = force_target.normalize(archive, logger)
        assert is_reached is True

    def test_rms_with_units_succeeds(self, archive, logger):
        """Test that rms convergence works with dimensional units."""
        force_target = ForceConvergenceTarget()
        force_target.threshold = 1e-8 * ureg.newton  # Has units - valid for rms
        force_target.threshold_type = 'rms'

        scf_steps = SCFSteps(
            delta_force_abs=[1e-9, 5e-10, 2e-10] * ureg.newton  # Array for rms
        )
        outputs = Outputs(scf_steps=scf_steps)
        archive.data.outputs = [outputs]

        # Should validate and check convergence
        is_reached = force_target.normalize(archive, logger)
        assert is_reached is True



# Direct type serialization
original = m_float_bounded(dtype=float, bound=Bound('[0,1]'))
serialized = original.serialize_self()

# Reconstruction loses bounds information
from nomad.metainfo.data_type import normalize_type
reconstructed = normalize_type(serialized)
# Returns basic m_float64 without bounds!



import numpy as np
from nomad import utils
from nomad.datamodel import EntryArchive
from nomad.units import ureg

from nomad_simulations.schema_packages.atoms_state import AtomsState
from nomad_simulations.schema_packages.general import Simulation
from nomad_simulations.schema_packages.model_method import DFT
from nomad_simulations.schema_packages.model_system import ModelSystem
from nomad_simulations.schema_packages.numerical_settings import SelfConsistency
from nomad_simulations.schema_packages.outputs import Outputs, SCFSteps
from nomad_simulations.schema_packages.properties import TotalEnergy

logger = utils.get_logger(__name__)

simulation = Simulation()

model_system = ModelSystem(is_representative=True)
model_system.lattice_vectors = np.eye(3) * 5.0 * ureg.angstrom
model_system.periodic_boundary_conditions = [True, True, True]
model_system.positions = np.array([[0, 0, 0], [1.25, 1.25, 1.25]]) * ureg.angstrom
model_system.particle_states.append(AtomsState(chemical_symbol='Si'))
model_system.particle_states.append(AtomsState(chemical_symbol='Si'))
simulation.model_system.append(model_system)

method = DFT(
    name='DFT',
    type='KS',
    numerical_settings=[
        SelfConsistency(threshold_change=1e-6, threshold_change_unit='joule')
    ],
)
simulation.model_method.append(method)

outputs = Outputs(scf_steps=SCFSteps())
for value in [1.0, 1.5, 2.0, 2.1, 2.101]:
    outputs.total_energies.append(TotalEnergy(value=value * ureg.eV))
simulation.outputs.append(outputs)

outputs.normalize(archive=EntryArchive(), logger=logger)

assert outputs.model_system_ref is simulation.model_system[0]
assert outputs.model_method_ref is simulation.model_method[0]
assert outputs.scf_steps is not None
assert outputs.scf_steps.delta_energies_total is not None
assert len(outputs.scf_steps.delta_energies_total) == 4
first_delta_joule = outputs.scf_steps.delta_energies_total[0].to('joule').magnitude
assert np.isclose(first_delta_joule, (0.5 * ureg.eV).to('joule').magnitude)



from collections.abc import Callable

import numpy as np
import pytest
from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.plot import PlotlyFigure
from nomad.metainfo import Quantity
from plotly.graph_objects import Figure

from nomad_simulations.schema_packages.physical_property import PhysicalProperty

from . import logger


class DummyPhysicalProperty(PhysicalProperty):
    value = Quantity(
        type=np.float64,
        unit='eV',
        shape=['*', '*', '*', '*'],
        description="""
        This value is defined in order to test the functionality in `PhysicalProperty`.
        """,
    )

    def plot(self, **kwargs) -> list[PlotlyFigure]:
        """Test implementation of plot method."""
        fig = Figure()
        fig.add_scatter(x=[1, 2, 3], y=[1, 4, 2], name='test')
        plotly_figure = PlotlyFigure(label='test', figure=fig.to_plotly_json())
        return [plotly_figure]


class TestPhysicalProperty:
    """
    Test the `PhysicalProperty` class defined in `physical_property.py`.
    """

    def test_is_derived(self):
        """
        Test the `normalize` and `_is_derived` methods.
        """
        # Testing a directly parsed physical property
        not_derived_physical_property = PhysicalProperty()
        assert not_derived_physical_property._is_derived() is False
        not_derived_physical_property.normalize(EntryArchive(), logger)
        assert not_derived_physical_property.is_derived is False
        # Testing a derived physical property
        derived_physical_property = PhysicalProperty(
            physical_property_ref=not_derived_physical_property,
        )
        assert derived_physical_property._is_derived() is True
        derived_physical_property.normalize(EntryArchive(), logger)
        assert derived_physical_property.is_derived is True

    def test_normalization_flag(self):
        """
        Test that the normalization flag prevents duplicate normalization.
        """
        property_obj = DummyPhysicalProperty()

        # First normalization
        property_obj.normalize(EntryArchive(), logger)
        assert property_obj.m_cache.get('_is_normalized', False) is True

        # Store original figures count
        original_figures_count = len(property_obj.figures)

        # Second normalization should not duplicate work
        property_obj.normalize(EntryArchive(), logger)

        # Should still be marked as normalized
        assert property_obj.m_cache.get('_is_normalized', False) is True
        # Should not have duplicated figures
        assert len(property_obj.figures) == original_figures_count

    def test_plotting_and_contributions(self):
        """
        Test plotting integration and contributions normalization.
        """
        # Test main property plotting
        property_obj = DummyPhysicalProperty()
        property_obj.normalize(EntryArchive(), logger)

        assert len(property_obj.figures) > 0
        assert isinstance(property_obj.figures[0], PlotlyFigure)

        # Test contributions
        main_property = DummyPhysicalProperty()
        contribution = DummyPhysicalProperty(name='contribution')
        main_property.contributions = [contribution]
        main_property.normalize(EntryArchive(), logger)

        # Both should be normalized
        assert main_property.m_cache.get('_is_normalized', False) is True
        assert contribution.m_cache.get('_is_normalized', False) is True

    @pytest.mark.parametrize(
        'instantiator, reference',
        [
            (PhysicalProperty, 'PhysicalProperty'),
            (DummyPhysicalProperty, 'DummyPhysicalProperty'),
        ],
    )
    def test_name_setting_during_normalization(
        self, instantiator: Callable, reference: str
    ):
        """
        Test that the name is set during normalization for PhysicalProperty.
        """
        property_obj = instantiator()
        property_obj.normalize(EntryArchive(), logger)
        assert property_obj.name == reference

    @pytest.mark.parametrize(
        'has_nested_contributions, log_ref',
        [(True, True), (False, False)],
    )
    def test_contributions_validation(
        self, caplog, has_nested_contributions: bool, log_ref: bool
    ):
        """
        Test contributions validation during normalization.

        Args:
            has_nested_contributions: Whether to create nested contribution structure
            log_ref: Whether validation error should be logged
        """
        main_property = DummyPhysicalProperty(name='main')

        if has_nested_contributions:
            nested_contribution = DummyPhysicalProperty(name='nested')
            contribution_with_nested = DummyPhysicalProperty(name='parent_contribution')
            contribution_with_nested.contributions = [nested_contribution]
            main_property.contributions = [contribution_with_nested]
        else:
            contribution1 = DummyPhysicalProperty(name='contrib1')
            contribution2 = DummyPhysicalProperty(name='contrib2')
            main_property.contributions = [contribution1, contribution2]

        with caplog.at_level('ERROR'):
            main_property.normalize(EntryArchive(), logger)

        has_nested_error = any(
            'nested contributions' in record.message.lower()
            for record in caplog.records
        )
        assert has_nested_error == log_ref

        if log_ref:
            assert any('Contribution 0' in record.message for record in caplog.records)

    @pytest.mark.parametrize(
        'set_contribution_type_on_main, set_contribution_type_on_contrib, log_ref',
        [
            (False, False, False),
            (False, True, False),
            (True, False, True),
            (True, True, True),
        ],
    )
    def test_contribution_type_validation(
        self,
        caplog,
        set_contribution_type_on_main: bool,
        set_contribution_type_on_contrib: bool,
        log_ref: bool,
    ):
        """
        Test contribution_type validation during normalization.

        Args:
            set_contribution_type_on_main: Whether to set contribution_type on main property
            set_contribution_type_on_contrib: Whether to set contribution_type on contribution
            log_ref: Whether validation error should be logged
        """
        main_property = DummyPhysicalProperty(name='main')
        if set_contribution_type_on_main:
            main_property.contribution_type = 'invalid_main_type'

        contribution = DummyPhysicalProperty(name='contrib')
        if set_contribution_type_on_contrib:
            contribution.contribution_type = 'valid_contrib_type'

        main_property.contributions = [contribution]

        with caplog.at_level('ERROR'):
            main_property.normalize(EntryArchive(), logger)

        has_contrib_type_error = any(
            'contribution_type set but is not a contribution' in record.message.lower()
            for record in caplog.records
        )
        assert has_contrib_type_error == log_ref

    def test_is_contribution_method(self):
        """
        Test the _is_contribution helper method.
        """
        main_property = DummyPhysicalProperty(name='main')
        contribution = DummyPhysicalProperty(name='contrib')

        assert not main_property._is_contribution()
        assert not contribution._is_contribution()

        main_property.contributions = [contribution]
        main_property.normalize(EntryArchive(), logger)

        assert not main_property._is_contribution()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
from typing import Any

import numpy as np
from nomad.metainfo.data_type import ExactNumber, InexactNumber

# Match patterns like '[0,3)', '(0,5]', '[1,)', '(,10)', etc.
bounds_patt = re.compile(r'^([\[\(])(-?\d*\.?\d*|),\s*(-?\d*\.?\d*|)([\]\)])$')


def _flatten_values(data: Any) -> list[Any]:
    """Returns a list of all scalar values from nested list/array structure."""
    if isinstance(data, np.ndarray):
        return data.flatten().tolist()
    elif isinstance(data, list):
        result = []
        for item in data:
            if isinstance(item, list):
                result.extend(_flatten_values(item))
            else:
                result.append(item)
        return result
    else:
        return [data]


class Bound:
    """
    Bounds checker for numeric values using mathematical interval notation.
    `None` and `NaN` values are allowed and will simply pass the checks.

    Range specification:
        - '[0,1]': Closed interval, 0 ≤ x ≤ 1
        - '(0,1)': Open interval, 0 < x < 1
        - '[0,1)': Half-open interval, 0 ≤ x < 1
        - '[1,)': Lower bounded, x ≥ 1
        - '(,10]': Upper bounded, x ≤ 10
        - '': Unbounded (-∞, ∞)
    """

    __slots__ = (
        '_min_value',
        '_max_value',
        '_min_inclusive',
        '_max_inclusive',
        '_original_min_str',
        '_original_max_str',
    )

    def __init__(self, range_str: str = ''):
        """Initialize bounds from range string.

        Args:
            range_str: Range specification like '[0,1]', '(0,)', etc. Empty means unbounded.
        """
        min_val, max_val, min_inc, max_inc, min_str, max_str = self._parse_range(
            range_str
        )
        self._min_value = min_val
        self._max_value = max_val
        self._min_inclusive = min_inc
        self._max_inclusive = max_inc
        self._original_min_str = min_str
        self._original_max_str = max_str

    def _parse_range(self, range_str: str) -> tuple[float, float, bool, bool, str, str]:
        """Parse range string like '[0,3)' into (min_val, max_val, min_inc, max_inc, min_str, max_str)."""
        if not range_str.strip():
            return float('-inf'), float('inf'), False, False, '', ''

        match = bounds_patt.match(range_str.strip())

        if not match:
            raise ValueError(
                f"Invalid range format: '{range_str}'. "
                f"Expected format like '[0,3)', '(0,5]', '[1,)', '(,10)', etc."
            )

        left_bracket, min_str, max_str, right_bracket = match.groups()

        # Parse bounds (empty means infinity)
        min_val = float('-inf') if not min_str else float(min_str)
        max_val = float('inf') if not max_str else float(max_str)

        # Parse inclusivity
        min_inclusive = left_bracket == '[' and bool(min_str)
        max_inclusive = right_bracket == ']' and bool(max_str)

        return min_val, max_val, min_inclusive, max_inclusive, min_str, max_str

    def _check_single_value(self, value: int | float) -> bool:
        """Check if a single value is within the specified bounds."""
        # lower bound
        if np.isfinite(self._min_value):
            if self._min_inclusive:
                if value < self._min_value:
                    return False
            else:
                if value <= self._min_value:
                    return False

        # upper bound
        if np.isfinite(self._max_value):
            if self._max_inclusive:
                if value > self._max_value:
                    return False
            else:
                if value >= self._max_value:
                    return False

        return True

    def check(self, value: Any, **kwargs) -> Any:
        """Check if value(s) are within bounds. Handles both scalar and array values.

        Note: NaN values will pass bounds checking since NaN comparisons always return False.

        Args:
            value: Value or array to check
            **kwargs: Additional arguments (for compatibility)

        Returns:
            The input value if valid

        Raises:
            ValueError: If any values are outside the bounds
        """
        if value is None:
            return value

        if flat_values := _flatten_values(value):
            invalid_values = [v for v in flat_values if not self._check_single_value(v)]

            if invalid_values:
                min_val = min(flat_values)
                max_val = max(flat_values)
                raise ValueError(
                    f'All values must be in {self}, got range [{min_val}, {max_val}]'
                )

        return value

    def __repr__(self) -> str:
        """Get string representation of bounds."""
        left = '[' if self._min_inclusive else '('
        right = ']' if self._max_inclusive else ')'

        min_str = self._original_min_str if np.isfinite(self._min_value) else ''
        max_str = self._original_max_str if np.isfinite(self._max_value) else ''

        return f'{left}{min_str},{max_str}{right}'


class m_int_bounded(ExactNumber):
    """
    Bounded integer data type.

    Example:
        m_int_bounded(dtype=int, bound=Bound('[1,10]'))    # 1 ≤ x ≤ 10 (integers)
    """

    __slots__ = ('bound',)

    def __init__(self, dtype=int, bound=None):
        """Initialize bounded integer with dtype and bounds.

        Args:
            dtype: Integer data type, mostly used to specify framework and accuracy (int, np.int32, etc.)
            bound: Bound instance specifying the valid range
        """
        super().__init__(dtype)
        self.bound = bound or Bound()

    def convertible_from(self, other):
        """Check if this data type can convert from another type."""
        # Follow the same convertibility rules as the base dtype
        if self._dtype in {int, np.int64}:
            return other in (int, np.int64, np.int32, np.int16, np.int8)
        elif self._dtype is np.int32:
            return other in (np.int32, np.int16, np.int8)
        elif self._dtype is np.int16:
            return other in (np.int16, np.int8)
        elif self._dtype is np.int8:
            return other is np.int8
        else:
            return False

    def serialize_self(self):
        """Serialize the datatype configuration."""
        return super().serialize_self() | {'type_bound': str(self.bound)}

    def normalize_flags(self, flags: dict):
        """Reconstruct from serialized data."""
        bounds_str = flags.get('type_bound', '')
        self.bound = Bound(bounds_str)
        # Apply any flags to base datatype
        super().normalize_flags(flags)
        return self

    def normalize(self, value, **kwargs):
        """Normalize value and validate bounds."""
        normalized_value = super().normalize(value, **kwargs)
        return self.bound.check(normalized_value, **kwargs)

    def standard_type(self):
        """Return the equivalent python type for indexing."""
        return 'int'


class m_float_bounded(InexactNumber):
    """
    Bounded float data type.

    Example:
        m_float_bounded(dtype=float, bound=Bound('[0.0,1.0]'))    # 0.0 ≤ x ≤ 1.0 (floats)
    """

    __slots__ = ('bound',)

    def __init__(self, dtype=float, bound=None):
        """Initialize bounded float with dtype and bounds.

        Args:
            dtype: Float data type, mostly used to specify framework and accuracy (float, np.float64, etc.)
            bound: Bound instance specifying the valid range
        """
        super().__init__(dtype)
        self.bound = bound or Bound()

    def convertible_from(self, other):
        """Check if this data type can convert from another type."""
        # Follow the same convertibility rules as the base dtype
        if self._dtype in {float, np.float64}:
            return other in (float, np.float64, np.float32, np.float16)
        elif self._dtype is np.float32:
            return other in (np.float32, np.float16)
        elif self._dtype is np.float16:
            return other is np.float16
        else:
            return False

    def serialize_self(self):
        """Serialize the datatype configuration."""
        return super().serialize_self() | {'type_bound': str(self.bound)}

    def normalize_flags(self, flags: dict):
        """Reconstruct from serialized data."""
        bounds_str = flags.get('type_bound', '')
        self.bound = Bound(bounds_str)
        # Apply any flags to base datatype
        super().normalize_flags(flags)
        return self

    def normalize(self, value, **kwargs):
        """Normalize value and validate bounds."""
        normalized_value = super().normalize(value, **kwargs)
        return self.bound.check(normalized_value, **kwargs)

    def standard_type(self):
        """Return the equivalent python type for indexing."""
        return 'float'


# Convenience factory functions for common use cases
def strictly_positive_int(*, dtype=int) -> m_int_bounded:
    """Create strictly positive integer type (x ≥ 1)."""
    return m_int_bounded(dtype=dtype, bound=Bound('[1,)'))


def positive_int(*, dtype=int) -> m_int_bounded:
    """Create positive integer type (x ≥ 0)."""
    return m_int_bounded(dtype=dtype, bound=Bound('[0,)'))


def strictly_positive_float(*, dtype=float) -> m_float_bounded:
    """Create strictly positive float type (x > 0)."""
    return m_float_bounded(dtype=dtype, bound=Bound('(0,)'))


def positive_float(*, dtype=float) -> m_float_bounded:
    """Create positive float type (x ≥ 0)."""
    return m_float_bounded(dtype=dtype, bound=Bound('[0,)'))


def unit_float(*, dtype=float) -> m_float_bounded:
    """Create unit interval float type (0 ≤ x ≤ 1)."""
    return m_float_bounded(dtype=dtype, bound=Bound('[0,1]'))



# Create interface system
interface = ModelSystem(is_representative=True)
interface.lattice_vectors = ...
interface.positions = ...

# Add particle states for both materials
for symbol in material_A_symbols + material_B_symbols:
    atom = AtomsState(chemical_symbol=symbol)
    interface.particle_states.append(atom)

# Define material A region as sub-system
material_A = ModelSystem(
    type='region',
    branch_label='Material A',
    particle_indices=list(range(len(material_A_symbols)))
)
interface.sub_systems.append(material_A)

# Define material B region as sub-system
material_B = ModelSystem(
    type='region',
    branch_label='Material B',
    particle_indices=list(range(len(material_A_symbols),
                                len(material_A_symbols) + len(material_B_symbols)))
)
interface.sub_systems.append(material_B)



import pytest

from nomad_simulations.schema_packages.model_method import ModelMethod
from nomad_simulations.schema_packages.model_system import ModelSystem
from nomad_simulations.schema_packages.outputs import Outputs
from nomad_simulations.schema_packages.workflow.single_point import (
    SinglePoint,
    SinglePointMethod,
    SinglePointResults,
)


class TestSinglePoint:
    def test_inputs_outputs(self, logger, archive):
        archive.data.outputs = [Outputs()]
        archive.data.model_method = [ModelMethod()]
        archive.data.model_system = [ModelSystem()]
        workflow = SinglePoint()
        workflow.normalize(archive, logger)
        assert isinstance(workflow.method, SinglePointMethod)
        assert isinstance(workflow.results, SinglePointResults)
        assert len(workflow.inputs) == 1
        assert workflow.inputs[0].name == 'Single point model'
        assert workflow.inputs[0].section.initial_method == archive.data.model_method[0]
        assert workflow.outputs[0].name == 'Single point results'
        assert len(workflow.outputs) == 2
        assert workflow.outputs[1].section == archive.data.outputs[0]

    @pytest.mark.parametrize('n_outputs', [(1), (2)])
    def test_tasks(self, logger, archive, log_output, n_outputs):
        archive.data.outputs = [Outputs() for _ in range(n_outputs)]
        archive.data.model_system = [ModelSystem()]
        workflow = SinglePoint()
        workflow.normalize(archive, logger)

        if n_outputs > 1:
            assert 'Incorrect number of tasks found.' in log_output.entries[0]['event']
        else:
            assert len(workflow.tasks) == 1
            assert workflow.tasks[0].name == 'Calculation'
            assert workflow.method in [inp.section for inp in workflow.tasks[0].inputs]
            assert archive.data.outputs[0] in [
                out.section for out in workflow.tasks[0].outputs
            ]



# Closed intervals (inclusive bounds)
Bound('[0,1]')      # 0 ≤ x ≤ 1
Bound('[1,10]')     # 1 ≤ x ≤ 10

# Open intervals (exclusive bounds)  
Bound('(0,1)')      # 0 < x < 1
Bound('(-1,1)')     # -1 < x < 1

# Half-open intervals
Bound('[0,1)')      # 0 ≤ x < 1
Bound('(0,1]')      # 0 < x ≤ 1

# Unbounded intervals
Bound('[0,)')       # x ≥ 0 (non-negative)
Bound('(0,)')       # x > 0 (strictly positive)
Bound('(,10]')      # x ≤ 10 (upper bounded)
Bound('(,-1)')      # x < -1 (strictly negative)

# Unbounded (no constraints)
Bound('')           # No bounds (-∞, ∞)



from typing import Optional

import pytest

from nomad_simulations.schema_packages.properties import FermiSurface


class TestFermiSurface:
    """
    Test the `FermiSurface` class defined in `properties/band_structure.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    @pytest.mark.parametrize(
        'n_bands',
        [
            (None),
            (10),
        ],
    )
    def test_default_quantities(self, n_bands: int | None):
        """
        Test the default quantities assigned when creating an instance of the `HoppingMatrix` class.
        """
        fermi_surface = FermiSurface(n_bands=n_bands)
        assert fermi_surface.iri == 'http://fairmat-nfdi.eu/taxonomy/FermiSurface'



import numpy as np
import pytest
from nomad.datamodel import EntryArchive
from nomad.units import ureg

from nomad_simulations.schema_packages.physical_property import PhysicalProperty
from nomad_simulations.schema_packages.properties.energies import (
    BaseEnergy,
    KineticEnergy,
    PotentialEnergy,
    TotalEnergy,
)

from . import logger


def test_energy_names_and_units_via_values():
    kin = KineticEnergy(value=1.23 * ureg.joule)
    pot = PotentialEnergy(value=4.56 * ureg.joule)
    tot = TotalEnergy(value=5.79 * ureg.joule)

    for sec in (kin, pot, tot):
        sec.normalize(EntryArchive(), logger)
        assert sec.name == sec.__class__.__name__
        assert hasattr(sec.value, 'magnitude')
        _ = sec.value.to('joule')
        assert np.isfinite(sec.value.magnitude)


def test_total_energy_contributions_basic_flow_and_idempotency():
    kin = KineticEnergy(value=2.0 * ureg.joule)
    pot = PotentialEnergy(value=3.0 * ureg.joule)

    tot = TotalEnergy(value=5.0 * ureg.joule)
    tot.contributions = [kin, pot]

    assert not tot.m_cache.get('_is_normalized', False)

    tot.normalize(EntryArchive(), logger)
    assert tot.m_cache.get('_is_normalized', False)
    assert len(tot.contributions) == 2

    assert kin._is_contribution() is True
    assert pot._is_contribution() is True
    assert tot._is_contribution() is False

    figures_before = len(tot.figures)
    tot.normalize(EntryArchive(), logger)
    assert len(tot.figures) == figures_before


def test_total_energy_accepts_baseenergy_contributions_mixed():
    """
    TotalEnergy.contributions accepts BaseEnergy directly, and mixed subclasses.
    """
    gen = BaseEnergy(value=1.0 * ureg.joule, contribution_type='generic')
    kin = KineticEnergy(value=2.0 * ureg.joule)  # subclass of BaseEnergy
    pot = PotentialEnergy(value=3.0 * ureg.joule)  # subclass of BaseEnergy

    tot = TotalEnergy(value=6.0 * ureg.joule)
    tot.contributions = [gen, kin, pot]

    tot.normalize(EntryArchive(), logger)

    # Top-level is not a contribution; children are.
    assert tot._is_contribution() is False
    assert all(c._is_contribution() for c in tot.contributions)

    # Names and units behave as expected
    assert gen.name == 'BaseEnergy'
    for c in tot.contributions:
        _ = c.value.to('joule')



from nomad.units import ureg

from nomad_simulations.schema_packages.outputs import Outputs
from nomad_simulations.schema_packages.properties import ElectronicBandGap

outputs = Outputs()
outputs.electronic_band_gaps.append(ElectronicBandGap(value=2.0 * ureg.eV))
outputs.electronic_band_gaps.append(
    ElectronicBandGap(value=1.8 * ureg.eV, spin_channel=0)
)
outputs.electronic_band_gaps.append(
    ElectronicBandGap(value=1.9 * ureg.eV, spin_channel=1)
)

spin_polarized = outputs.extract_spin_polarized_property('electronic_band_gaps')

assert len(outputs.electronic_band_gaps) == 3
assert len(spin_polarized) == 2
assert {gap.spin_channel for gap in spin_polarized} == {0, 1}



from typing import Optional

import numpy as np
import pytest

from nomad_simulations.schema_packages.properties import Permittivity
from nomad_simulations.schema_packages.variables import Frequency, KMesh, Variables

from ..conftest import generate_k_space_simulation
from . import logger


class TestPermittivity:
    """
    Test the `Permittivity` class defined in `properties/permittivity.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    def test_default_quantities(self):
        """
        Test the default quantities assigned when creating an instance of the `Permittivity` class.
        """
        permittivity = Permittivity()
        assert permittivity.iri == 'http://fairmat-nfdi.eu/taxonomy/Permittivity'

    @pytest.mark.parametrize(
        'kmesh_grid, frequency, result',
        [
            (None, None, 'static'),
            ([], None, 'static'),
            ([3, 3, 3], None, 'static'),
            ([3, 3, 3], Frequency(), 'dynamic'),
        ],
    )
    def test_resolve_type(
        self,
        kmesh_grid: list[int] | None,
        frequency: Variables | None,
        result: str,
    ):
        """
        Test the `resolve_type` method.
        """
        permittivity = Permittivity()

        # Generating `KMesh` numerical settings section
        simulation = generate_k_space_simulation(grid=kmesh_grid)
        k_mesh_settings = simulation.model_method[0].numerical_settings[0].k_mesh[0]
        k_mesh_settings.label = 'q-mesh'
        k_mesh_settings.center = 'Monkhorst-Pack'
        k_mesh_settings.points, _ = k_mesh_settings.resolve_points_and_offset(logger)
        if kmesh_grid is not None and len(kmesh_grid) > 0:
            permittivity.q_mesh = KMesh(points=k_mesh_settings)
        if frequency is not None:
            permittivity.frequencies = frequency
        assert permittivity.resolve_type() == result

    @pytest.mark.parametrize(
        'kmesh_grid, frequencies_points, value, result',
        [
            # Empty case
            (None, None, None, None),
            # No `variables`
            ([], [], np.eye(3) * (1 + 1j), None),
            # If `KMesh` is defined we cannot extract absorption spectra
            (
                [4, 1, 1],
                [],
                np.array([np.eye(3) * k_point * (1 + 1j) for k_point in range(1, 5)]),
                None,
            ),
            # Even if we define `Frequency`, we cannot extract absorption spectra if `value` depends on `KMesh`
            (
                [4, 1, 1],
                [0, 1, 2, 3, 4],
                np.array(
                    [
                        [
                            np.eye(3) * k_point * (1 + 1j)
                            + np.eye(3) * freq_point * 0.5j
                            for freq_point in range(5)
                        ]
                        for k_point in range(1, 5)
                    ]
                ),
                None,
            ),
            # Valid case: `value` does not depend on `KMesh` and we can extract absorption spectra
            (
                [],
                [0, 1, 2, 3, 4],
                np.array([np.eye(3) * freq_point * 0.5j for freq_point in range(5)]),
                [0.0, 0.5, 1.0, 1.5, 2.0],
            ),
        ],
    )
    def test_extract_absorption_spectra(
        self,
        kmesh_grid: list[int] | None,
        frequencies_points: list[float] | None,
        value: np.ndarray | None,
        result: list[float] | None,
    ):
        """
        Test the `extract_absorption_spectra` method. The `result` in the last valid case corresponds to the imaginary part of
        the diagonal of the `Permittivity.value` for each frequency point.
        """
        permittivity = Permittivity()

        # Generating `KMesh` numerical settings section
        simulation = generate_k_space_simulation(grid=kmesh_grid)
        k_mesh_settings = simulation.model_method[0].numerical_settings[0].k_mesh[0]
        k_mesh_settings.label = 'q-mesh'
        k_mesh_settings.center = 'Monkhorst-Pack'
        k_mesh_settings.points, _ = k_mesh_settings.resolve_points_and_offset(logger)
        if kmesh_grid is not None and len(kmesh_grid) > 0:
            kmesh_variables = KMesh(points=k_mesh_settings)
            permittivity.q_mesh = kmesh_variables

        # Adding `Frequency` if defined
        if frequencies_points is not None and len(frequencies_points) > 0:
            frequencies = Frequency(points=frequencies_points)
            permittivity.frequencies = frequencies

        if permittivity.q_mesh is not None or permittivity.frequencies is not None:
            permittivity.value = value

        absorption_spectra = permittivity.extract_absorption_spectra(logger)
        if absorption_spectra is not None:
            assert len(absorption_spectra) == 3
            spectrum = absorption_spectra[1]
            assert spectrum.axis == 'yy'
            if spectrum.frequencies is not None:
                assert len(spectrum.value) == len(permittivity.frequencies.points)
            assert np.allclose(spectrum.value, result)
        else:
            assert absorption_spectra == result



from nomad_simulations.schema_packages.workflow.single_point import SinglePoint
from nomad_simulations.schema_packages.workflow.tb import (
    DFTTBMethod,
    DFTTBResults,
    DFTTBWorkflow,
)


class TestDFTTBWorkflow:
    def test_inputs_outputs(self, archive, logger, log_output):
        workflow = DFTTBWorkflow()
        workflow.normalize(archive, logger)
        assert isinstance(workflow.method, DFTTBMethod)
        assert isinstance(workflow.results, DFTTBResults)
        assert len(workflow.inputs) == 1
        assert len(workflow.outputs) == 1
        assert workflow.inputs[0].name == 'DFT+TB workflow parameters'
        assert workflow.outputs[0].name == 'DFT+TB worklfow results'

    def test_tasks(self, archive, logger):
        workflow = DFTTBWorkflow(tasks=[SinglePoint(name='GS'), SinglePoint()])
        workflow.normalize(archive, logger)
        assert workflow.tasks[1].name == 'TB'
        assert workflow.tasks[0] in [inp.section for inp in workflow.tasks[1].inputs]



from nomad.metainfo import Quantity, Section
from nomad_simulations.schema_packages.data_types import (
    Bound, m_int_bounded, m_float_bounded
)

class MySection(Section):
    # Integer value constrained to [1, 10]
    count = Quantity(
        type=m_int_bounded(dtype=int, bound=Bound('[1,10]')),
        description='Number of items (1-10)'
    )
    
    # Float value constrained to [0.0, 1.0]
    probability = Quantity(
        type=m_float_bounded(dtype=float, bound=Bound('[0.0,1.0]')),
        description='Probability value (0.0-1.0)'
    )
    
    # Array of positive floats
    energies = Quantity(
        type=m_float_bounded(dtype=float, bound=Bound('(0,)')),
        shape=['*'],
        description='Energy values (strictly positive)'
    )



import os
import shutil

from nomad import datamodel, utils
from nomad.files import StagingUploadFiles, UploadFiles


# TODO remove this only a temporary fix: function not included in package
def create_test_upload_files(
    upload_id: str | None,
    archives: list[datamodel.EntryArchive] | None = None,
    published: bool = True,
    embargo_length: int = 0,
    raw_files: str = None,
    template_files: str = '',
    template_mainfile: str = '',
    additional_files_path: str = None,
) -> UploadFiles:
    """
    Creates an upload_files object and the underlying files for test/mock purposes.

    Arguments:
        upload_id: The upload id for the upload. Will generate a random UUID if None.
        archives: A list of class:`datamodel.EntryArchive` metainfo objects. This will
            be used to determine the mainfiles. Will create respective directories and
            copy the template entry to create raw files for each archive.
            Will also be used to fill the archives in the create upload.
        published: Creates a :class:`PublicUploadFiles` object with published files
            instead of a :class:`StagingUploadFiles` object with staging files. Default
            is published.
        embargo_length: The embargo length
        raw_files: A directory path. All files here will be copied into the raw files
            dir of the created upload files.
        template_files: A zip file with example files in it. One directory will be used
            as a template. It will be copied for each given archive.
        template_mainfile: Path of the template mainfile within the given template_files.
        additional_files_path: Path to additional files to add.
    """
    if upload_id is None:
        upload_id = utils.create_uuid()
    if archives is None:
        archives = []

    upload_files = StagingUploadFiles(upload_id, create=True)
    if raw_files:
        shutil.rmtree(upload_files._raw_dir.os_path)
        shutil.copytree(raw_files, upload_files._raw_dir.os_path)
    if template_files:
        upload_files.add_rawfiles(template_files)
    if additional_files_path:
        upload_files.add_rawfiles(additional_files_path)

    for archive in archives:
        mainfile = archive.metadata.mainfile
        assert mainfile is not None, (
            'Archives to create test upload must have a mainfile'
        )
        # create an archive "file" for each archive
        entry_id = archive.metadata.entry_id
        assert entry_id is not None, (
            'Archives to create test upload must have an entry_id'
        )
        upload_files.write_archive(entry_id, archive.m_to_dict())

    if published:
        upload_files.pack(
            [archive.metadata for archive in archives], with_embargo=embargo_length > 0
        )
        upload_files.delete()
        return UploadFiles.get(upload_id)

    return upload_files
